r"""
**********************
Single-variant testing
**********************

Linear models
^^^^^^^^^^^^^

.. autofunction:: limix.qtl.qtl_test_lm

Linear mixed models
^^^^^^^^^^^^^^^^^^^

.. autofunction:: limix.qtl.qtl_test_lmm
.. autofunction:: limix.qtl.qtl_test_interaction_lmm
.. autofunction:: limix.qtl.qtl_test_lmm_kronecker
.. autofunction:: limix.qtl.qtl_test_interaction_lmm_kronecker
.. autofunction:: limix.qtl.forward_lmm
.. autoclass:: limix.qtl.LMM
  :members:

Generalised linear mixed models
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autofunction:: limix.qtl.qtl_test_glmm

"""

from .qtl import qtl_test_lm
from .qtl import qtl_test_lmm
from .qtl import qtl_test_lmm_kronecker
from .qtl import qtl_test_interaction_lmm_kronecker
from .qtl import qtl_test_interaction_lmm
from .qtl import forward_lmm
from .glmm import qtl_test_glmm
from .lmm import LMM

__all__ = ['qtl_test_lm', 'qtl_test_lmm', 'qtl_test_lmm_kronecker',
           'qtl_test_interaction_lmm_kronecker', 'qtl_test_interaction_lmm',
           'forward_lmm', 'LMM', 'qtl_test_glmm']



import limix
import scipy as sp
from .linalg_utils import lowrank_approx

ntype_dict = {'assoc': 'null', 'gxe': 'block', 'gxehet': 'rank1'}


def define_gp(Y, Xr, F, type, Rr):
    from limix_core.covar import LowRankCov
    from limix_core.covar import FixedCov
    from limix_core.covar import FreeFormCov
    from limix_core.gp import GP2KronSumLR
    from limix_core.gp import GP2KronSum

    P = Y.shape[1]
    _A = sp.eye(P)
    if type in ['null', 'rank1']:
        _Cr = LowRankCov(P, 1)
    elif type == 'block':
        _Cr = FixedCov(sp.ones((P, P)))
    elif type == 'full':
        _Cr = FreeFormCov(P)
    else:
        print('poppo')
    _Cn = FreeFormCov(P)
    if type == 'null':
        _gp = GP2KronSumLR(
            Y=Y, G=sp.ones(
                (Y.shape[0], 1)), F=F, A=_A, Cr=_Cr, Cn=_Cn)
        _Cr.setParams(1e-9 * sp.ones(P))
        _gp.covar.act_Cr = False
    else:
        if Xr.shape[1] < Xr.shape[0]:
            _gp = GP2KronSumLR(Y=Y, G=Xr, F=F, A=_A, Cr=_Cr, Cn=_Cn)
        else:
            _gp = GP2KronSum(Y=Y, F=F, A=_A, R=Rr, Cg=_Cr, Cn=_Cn)
    return _gp


class MvSetTest():
    """
    Args:
        Y (ndarray):
            (`N`, `P`) phenotype matrix for `N` samples and `P` traits
        Xr (ndarray):
            (`N`, `S`) genotype values for `N` samples and `S` variants
            (defines the set component)
        factr (float):
            optimization paramenter that determines the accuracy of the solution
            (see scipy.optimize.fmin_l_bfgs_b for more details).
    """

    def __init__(
            self,
            Y=None,
            Xr=None,
            Rr=None,
            F=None,
            factr=1e7,
            debug=False):
        # avoid SVD failure by adding some jitter
        Xr += 2e-6 * (sp.rand(*Xr.shape) - 0.5)
        # make sure it is normalised
        Xr -= Xr.mean(0)
        Xr /= Xr.std(0)
        Xr /= sp.sqrt(Xr.shape[1])
        self.Y = Y
        self.F = F
        self.Xr = Xr
        self.covY = sp.cov(Y.T)
        self.factr = factr
        self.debug = debug
        self.gp = {}
        self.info = {}
        self.lowrank = Xr.shape[1] < Xr.shape[0]
        if Rr is not None:
            self.Rr = Rr
        if self.lowrank:
            self.Rr = None
        else:
            self.Rr = sp.dot(Xr, Xr.T)

    def assoc(self):
        # fit model
        for key in ['null', 'full']:
            if key not in list(self.gp.keys()):
                if self.debug:
                    print('.. dening %s' % key)
                self.gp[key] = define_gp(
                    self.Y, self.Xr, self.F, key, Rr=self.Rr)
                if self.debug:
                    print('.. fitting %s' % key)
                self.info[key] = self._fit(key, vc=True)
                # if key=='full':
                #    print self.info[key]['var'][0]
        return self.info['null']['LML'] - self.info['full']['LML']

    def gxe(self):
        # fit model
        for key in ['null', 'full', 'block']:
            if key not in list(self.gp.keys()):
                if self.debug:
                    print('.. defining %s' % key)
                self.gp[key] = define_gp(
                    self.Y, self.Xr, self.F, key, Rr=self.Rr)
                if self.debug:
                    print('.. fitting %s' % key)
                self.info[key] = self._fit(key, vc=True)
        return self.info['block']['LML'] - self.info['full']['LML']

    def gxehet(self):
        # fit model
        for key in ['null', 'full', 'rank1']:
            if key not in list(self.gp.keys()):
                if self.debug:
                    print('.. defining %s' % key)
                self.gp[key] = define_gp(
                    self.Y, self.Xr, self.F, key, Rr=self.Rr)
                if self.debug:
                    print('.. fitting %s' % key)
                self.info[key] = self._fit(key, vc=True)
        return self.info['rank1']['LML'] - self.info['full']['LML']

    def assoc_null(self, n_nulls=30):
        LLR0 = sp.zeros(n_nulls)
        for ni in range(n_nulls):
            idx_perms = sp.random.permutation(self.Y.shape[0])
            _Xr = self.Xr[idx_perms]
            if self.Rr is not None:
                _Rr = sp.dot(_Xr, _Xr.T)
            else:
                _Rr = None
            mvset0 = MvSetTest(Y=self.Y, F=self.F, Xr=_Xr, Rr=_Rr)
            LLR0[ni] = mvset0.assoc()
        return LLR0

    def gxe_null(self, n_nulls=30):
        LLR0 = sp.zeros(n_nulls)
        for ni in range(n_nulls):
            if self.lowrank:
                _Y = self.gp['block'].simulate_pheno()
            else:
                _Y = self.gp['block'].simulate_pheno(Rh=self.Xr)
            mvset0 = MvSetTest(Y=_Y, F=self.F, Xr=self.Xr, Rr=self.Rr)
            LLR0[ni] = mvset0.gxe()
        return LLR0

    def gxehet_null(self, n_nulls=30):
        LLR0 = sp.zeros(n_nulls)
        for ni in range(n_nulls):
            if self.lowrank:
                _Y = self.gp['rank1'].simulate_pheno()
            else:
                _Y = self.gp['rank1'].simulate_pheno(Rh=self.Xr)
            mvset0 = MvSetTest(Y=_Y, F=self.F, Xr=self.Xr, Rr=self.Rr)
            LLR0[ni] = mvset0.gxehet()
        return LLR0

    def _fit(self, type, vc=False):
        # 2. init
        if type == 'null':
            self.gp[type].covar.Cn.setCovariance(self.covY)
        elif type == 'full':
            Cn0_K = self.gp['null'].covar.Cn.K()
            # self.gp[type].covar.Cr.setCovariance(1e-4*sp.ones(self.covY.shape)+1e-4*sp.eye(self.covY.shape[0]))
            self.gp[type].covar.Cr.setCovariance(0.5 * Cn0_K)
            self.gp[type].covar.Cn.setCovariance(0.5 * Cn0_K)
        elif type == 'block':
            Crf_K = self.gp['full'].covar.Cr.K()
            Cnf_K = self.gp['full'].covar.Cn.K()
            self.gp[type].covar.Cr.scale = sp.mean(Crf_K)
            self.gp[type].covar.Cn.setCovariance(Cnf_K)
        elif type == 'rank1':
            Crf_K = self.gp['full'].covar.Cr.K()
            Cnf_K = self.gp['full'].covar.Cn.K()
            self.gp[type].covar.Cr.setCovariance(Crf_K)
            self.gp[type].covar.Cn.setCovariance(Cnf_K)
        else:
            print('poppo')
        self.gp[type].optimize(factr=self.factr, verbose=False)
        RV = {'Cr': self.gp[type].covar.Cr.K(),
              'Cn': self.gp[type].covar.Cn.K(),
              'B': self.gp[type].mean.B[0],
              'LML': sp.array([self.gp[type].LML()]),
              'LMLgrad': sp.array([sp.mean((self.gp[type].LML_grad()['covar'])**2)])}
        if vc:
            # tr(P CoR) = tr(C)tr(R) - tr(Ones C) tr(Ones R) / float(NP)
            #           = tr(C)tr(R) - C.sum() * R.sum() / float(NP)
            trRr = (self.Xr**2).sum()
            var_r = sp.trace(RV['Cr']) * trRr / float(self.Y.size - 1)
            var_c = sp.var(sp.dot(self.F, RV['B']))
            var_n = sp.trace(RV['Cn']) * self.Y.shape[0]
            var_n -= RV['Cn'].sum() / float(RV['Cn'].shape[0])
            var_n /= float(self.Y.size - 1)
            RV['var'] = sp.array([var_r, var_c, var_n])
            if 0 and self.Y.size < 5000:
                pdb.set_trace()
                Kr = sp.kron(RV['Cr'], sp.dot(self.Xr, self.Xr.T))
                Kn = sp.kron(RV['Cn'], sp.eye(self.Y.shape[0]))
                _var_r = sp.trace(Kr - Kr.mean(0)) / float(self.Y.size - 1)
                _var_n = sp.trace(Kn - Kn.mean(0)) / float(self.Y.size - 1)
                _var = sp.array([_var_r, var_c, _var_n])
                print(((_var - RV['var'])**2).mean())
            if type == 'full':
                # calculate within region vcs
                Cr_block = sp.mean(RV['Cr']) * sp.ones(RV['Cr'].shape)
                Cr_rank1 = lowrank_approx(RV['Cr'], rank=1)
                var_block = sp.trace(Cr_block) * trRr / float(self.Y.size - 1)
                var_rank1 = sp.trace(Cr_rank1) * trRr / float(self.Y.size - 1)
                RV['var_r'] = sp.array(
                    [var_block, var_rank1 - var_block, var_r - var_rank1])
        return RV



r"""
******************
Plotting utilities
******************

Manhattan
^^^^^^^^^

.. autofunction:: limix.plot.plot_manhattan

QQ plot
^^^^^^^

.. autofunction:: limix.plot.qqplot

Power plots
^^^^^^^^^^^

.. autofunction:: limix.plot.plot_power_curve

Kinship
^^^^^^^

.. autofunction:: limix.plot.plot_kinship

Normal distribution
^^^^^^^^^^^^^^^^^^^

.. autofunction:: limix.plot.plot_normal

"""

from .plot_normal import plot_normal
from .qqplot import qqplot
from .manhattan import plot_manhattan
from .kinship import plot_kinship
from .image import see_image
from .power import plot_power_curve

__all__ = ['plot_manhattan', 'qqplot', 'plot_normal', 'plot_kinship',
           'see_image', 'plot_power_curve']



def pca(X, ncomp):
    r"""Principal component analysis.

    Args:
        X (array_like): data.
        ncomp (int): number of components.

    Returns:
        dict: dict containing:
            - **components** (*array_like*):
              first components ordered by explained variance.
            - **explained_variance** (*array_like*):
              explained variance.
            - **explained_variance_ratio** (*array_like*):
              percentage of variance explained.

    Examples
    --------

        .. doctest::

            >>> from numpy.random import RandomState
            >>> from limix.stats import pca
            >>> from numpy import set_printoptions
            >>> set_printoptions(4)
            >>>
            >>> X = RandomState(1).randn(4, 5)
            >>> result = pca(X, ncomp=2)
            >>> print(result['components']) # doctest: +SKIP
            [[-0.7502  0.5835 -0.0797  0.1957 -0.2285]
             [ 0.4884  0.7227  0.0197 -0.4616 -0.1603]]
            >>> print(result['explained_variance']) # doctest: +SKIP
            [6.4466 0.5145]
            >>> print(result['explained_variance_ratio']) # doctest: +SKIP
            [0.9205 0.0735]
    """
    from sklearn.decomposition import PCA

    pca = PCA(n_components=ncomp)
    pca.fit(X)

    return dict(
        components=pca.components_,
        explained_variance=pca.explained_variance_,
        explained_variance_ratio=pca.explained_variance_ratio_)



from __future__ import division

from dask.array import from_array
from numpy import logical_not
from numpy.random import RandomState
from numpy.testing import assert_allclose, assert_equal

from limix.stats import indep_pairwise, maf


def test_preprocess_indep_pairwise():
    random = RandomState(0)

    X = random.randn(3, 100)

    head = [True, True, False, True, False]
    tail = [True, True, False, False]

    assert_equal(indep_pairwise(X, 4, 2, 0.5, verbose=False)[:5], head)
    assert_equal(indep_pairwise(X, 4, 2, 0.5, verbose=False)[-4:], tail)

    X = from_array(X, chunks=(2, 10))
    assert_equal(indep_pairwise(X, 4, 2, 0.5, verbose=False)[:5], head)
    assert_equal(indep_pairwise(X, 4, 2, 0.5, verbose=False)[-4:], tail)


def test_preprocess_maf():
    random = RandomState(0)

    X = random.randint(0, 3, size=(100, 10))
    assert_allclose(
        maf(X), [0.49, 0.49, 0.445, 0.495, 0.5, 0.45, 0.48, 0.48, 0.47, 0.435])



import os
import time


def entry_point():
    from limix_legacy.deprecated.io.conversion import LIMIX_converter

    infostring = "limix_conveter.py, Copyright(c) 2014, "
    tim = time.ctime(os.path.getmtime(__file__))
    infostring += "The LIMIX developers\nlast modified: %s" % tim
    print(infostring)

    runner = LIMIX_converter(infostring=infostring)
    (options, args) = runner.parse_args()
    runner.run()



from __future__ import division

from numpy import arange, asarray, cumsum, flipud, log10


def plot_manhattan(df,
                   alpha=None,
                   null_style=dict(alpha=0.1, color='DarkBlue'),
                   alt_style=dict(alpha=0.5, color='Orange'),
                   ax=None):
    r"""Produce a manhattan plot.

    Parameters
    ----------
    df : :class:`pandas.DataFrame`
        A Pandas DataFrame containing columns pv for p-values, pos for
        base-pair positions, and chrom for chromossome names..
    alpha : float
        Threshold for significance. Defaults to 0.01 significance level
        (bonferroni-adjusted).
    ax : :class:`matplotlib.axes.AxesSubplot`:
        The target handle for this figure. If None, the current axes is set.

    Returns
    -------
    :class:`matplotlib.axes.AxesSubplot`
        Axes object.

    Examples
    --------
    .. plot::

        from numpy.random import RandomState
        from numpy import arange, ones, kron
        import pandas as pd
        from limix.plot import plot_manhattan
        from matplotlib import pyplot as plt
        random = RandomState(1)
        pv = random.rand(5000)
        pv[1200:1250] = random.rand(50)**4
        chrom  = kron(arange(1,6), ones(1000))
        pos = kron(ones(5), arange(1,1001))
        data = dict(pv=pv, chrom=chrom, pos=pos)
        plot_manhattan(pd.DataFrame(data=data))
        plt.tight_layout()
        plt.show()
    """

    import matplotlib.pyplot as plt

    ax = plt.gca() if ax is None else ax

    if 'pos' not in df:
        df['pos'] = arange(df.shape[0])

    if 'label' not in df:
        chrom = df['chrom'].astype(int).astype(str)
        pos = df['pos'].astype(int).astype(str)
        df['label'] = (
            'chrom' + chrom + '_pos' + pos)

    df = _abs_pos(df)

    if alpha is None:
        alpha = 0.01 / df.shape[0]

    ytop = -1.2 * log10(min(df['pv'].min(), alpha))

    _plot_chrom_strips(ax, df, ytop)
    _plot_points(ax, df, alpha, null_style, alt_style)
    _set_frame(ax, df, ytop)

    ax.set_ylabel('-log$_{10}$pv')
    ax.set_xlabel('chromosome')

    _set_ticks(ax, _chrom_bounds(df))

    return ax


def _set_frame(ax, df, ytop):
    ax.set_ylim(0, ytop)
    ax.set_xlim(0, df['abs_pos'].max())

    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)


def _plot_points(ax, df, alpha, null_style, alt_style):
    null_df = df.loc[df['pv'] >= alpha, :]
    alt_df = df.loc[df['pv'] < alpha, :]

    ax.plot(null_df['abs_pos'], -log10(null_df['pv']), '.', ms=5, **null_style)
    ax.plot(alt_df['abs_pos'], -log10(alt_df['pv']), '.', ms=5, **alt_style)

    for i in range(alt_df.shape[0]):
        x = alt_df['abs_pos'].values[i]
        y = -log10(alt_df['pv'].values[i])
        _annotate(ax, x, y, alt_df['label'].values[i])


def _plot_chrom_strips(ax, df, ytop):
    uchroms = df['chrom'].unique()
    for i in range(0, len(uchroms), 2):
        ax.fill_between(
            df['abs_pos'],
            0,
            ytop,
            where=df['chrom'] == uchroms[i],
            facecolor='LightGray',
            linewidth=0,
            alpha=0.5)


def _set_ticks(ax, chrom_bounds):
    n = len(chrom_bounds) - 1
    xticks = asarray([chrom_bounds[i:i + 2].mean() for i in range(n)])
    ax.set_xticks(xticks)
    ax.tick_params(axis='x', which='both', labelsize=6)
    ax.set_xticklabels(arange(1, n + 2))
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')


def _abs_pos(df):
    uchroms = df['chrom'].unique()
    chrom_ends = [df['pos'][df['chrom'] == c].max() for c in uchroms]

    offset = flipud(cumsum(chrom_ends)[:-1])

    df['abs_pos'] = df['pos'].copy()

    uchroms = list(reversed(uchroms))
    for i in range(len(offset)):
        ix = df['chrom'] == uchroms[i]
        df.loc[ix, 'abs_pos'] = df.loc[ix, 'abs_pos'] + offset[i]

    return df


def _chrom_bounds(df):
    uchroms = df['chrom'].unique()
    v = [df['abs_pos'][df['chrom'] == c].min() for c in uchroms]
    return asarray(v + [df['abs_pos'].max()])


def _annotate(ax, x, y, text):
    ax.annotate(
        text,
        xy=(x, y),
        xytext=(-18, 18),
        textcoords='offset points',
        fontsize=6,
        ha='center',
        va='bottom',
        bbox=dict(boxstyle='round,pad=0.2', fc='yellow', alpha=0.3),
        arrowprops=dict(
            arrowstyle='->', connectionstyle='arc3,rad=0.5', color='red'))



from os.path import dirname, realpath, join

from numpy.testing import assert_allclose, assert_equal

from limix.io import h5data_fetcher, read_csv, read_gen


def test_io_hdf5():
    dir_path = dirname(realpath(__file__))
    with h5data_fetcher(join(dir_path, 'data', 'hdf5', 'data.h5')) as df:
        X = df.fetch('/genotype/chrom20/SNP')
        assert_allclose(X[0, 0].compute(), [1.0])


def test_io_csv():
    dir_path = dirname(realpath(__file__))
    fn = join(dir_path, 'data', 'csv', 'pheno.csv')
    assert_equal(read_csv(fn)['attr1'].compute()[0], 'string')


def test_io_gen():
    dir_path = dirname(realpath(__file__))
    df = read_gen(join(dir_path, 'data', 'gen', 'example'))
    assert_equal(df['sample']['sample_id'][1], '1A1')
    assert_equal(df['sample']['age'][0], 4)
    assert_allclose(df['genotype']['1A4']['AB'][0], 0.0207)



import sys
import h5py
import pdb
import scipy as SP
import scipy.stats as ST
import scipy.linalg as LA
import time as TIME
import copy
import warnings
import os
import csv


def splitGeno(
        pos,
        method='slidingWindow',
        size=5e4,
        step=None,
        annotation_file=None,
        cis=1e4,
        funct=None,
        out_file=None):
    """
    split geno into windows and store output in csv file
    Args:
        pos:    genomic position in the format (chrom,pos)
        method:     method used to slit the windows:
                    'slidingWindow':    uses a sliding window
                    'geneWindow':       uses windows centered on genes
        size:       window size used in slidingWindow method
        step:       moving step used in slidingWindow method
        annotation_file:    file containing the annotation file for geneWindow method
        out_file:   output csv file
    """
    assert method in ['slidingWindow', 'geneWindow'], 'method not known'

    # create folder if does not exists
    out_dir, fname = os.path.split(out_file)
    if (out_dir != '') and (not os.path.exists(out_dir)):
        os.makedirs(out_dir)

    # calculates windows using the indicated method
    if method == 'slidingWindow':
        nWnds, nSnps = splitGenoSlidingWindow(
            pos, out_file, size=size, step=step)
    elif method == 'geneWindow':
        # out = splitGenoGeneWindow(pos,out_file,annotation_file=annotation_file,cis=cis,funct=funct)
        pass

    return nWnds, nSnps


def splitGenoSlidingWindow(pos, out_file, size=5e4, step=None):
    """
    split into windows using a slide criterion
    Args:
        size:       window size
        step:       moving step (default: 0.5*size)
    Returns:
        wnd_i:      number of windows
        nSnps:      vector of per-window number of SNPs
    """
    if step is None:
        step = 0.5 * size
    chroms = SP.unique(pos[:, 0])

    RV = []
    wnd_i = 0
    wnd_file = csv.writer(open(out_file, 'w'), delimiter='\t')
    nSnps = []
    for chrom_i in chroms:
        Ichrom = pos[:, 0] == chrom_i
        idx_chrom_start = SP.where(Ichrom)[0][0]
        pos_chr = pos[Ichrom, 1]
        start = pos_chr.min()
        pos_chr_max = pos_chr.max()
        while True:
            if start > pos_chr_max:
                break
            end = start + size
            Ir = (pos_chr >= start) * (pos_chr < end)
            _nSnps = Ir.sum()
            if _nSnps > 0:
                idx_wnd_start = idx_chrom_start + SP.where(Ir)[0][0]
                nSnps.append(_nSnps)
                line = SP.array([wnd_i, chrom_i, start, end,
                                 idx_wnd_start, _nSnps], dtype=int)
                wnd_file.writerow(line)
                wnd_i += 1
            start += step
    nSnps = SP.array(nSnps)
    return wnd_i, nSnps


def splitGenoGeneWindow(
        self,
        annotation_file=None,
        cis=1e4,
        funct='protein_coding'):
    """
    split into windows based on genes
    """
    # 1. load annotation
    assert annotation_file is not None, 'Splitter:: specify annotation file'
    try:
        f = h5py.File(annotation_file, 'r')
        geneID = f['geneID'][:]
        gene_chrom = f['chrom'][:]
        gene_start = f['start'][:]
        gene_end = f['end'][:]
        gene_strand = f['strand'][:]
        gene_function = f['function'][:]
        f.close()
    except BaseException:
        print('Splitter:: format annotation file not valid')

    # if funct is not None, it has to be a list
    if funct is not None and funct != list:
        funct = [funct]

    windows = []
    nSnps = []
    Igene = []
    # 2. calculates windows
    for gene_i in range(geneID.shape[0]):
        if funct is not None:
            if gene_function[gene_i] not in funct:
                Igene.append(False)
                continue
        wnd = [
            gene_chrom[gene_i],
            gene_start[gene_i] - cis,
            gene_end[gene_i] + cis]
        Ir = (self.chrom == wnd[0]) * \
            (self.pos >= wnd[1]) * (self.pos <= wnd[2])
        _nSnps = Ir.sum()
        if _nSnps >= minSnps and _nSnps <= maxSnps:
            windows.append(wnd)
            nSnps.append(_nSnps)
            Igene.append(True)
        else:
            Igene.append(False)
    Igene = SP.array(Igene)
    self.info['nSnps'] = SP.array(nSnps)
    self.info['geneID'] = geneID[Igene]
    self.info['gene_start'] = gene_start[Igene]
    self.info['gene_end'] = gene_end[Igene]
    self.info['gene_strand'] = gene_strand[Igene]
    self.info['gene_function'] = gene_function[Igene]
    return SP.array(windows)


if __name__ == "__main__":

    data = './../data/1000G_chr22/chrom22'
    window_size = 1e4
    precompute_windows(data, size=window_size, plot=True)



def read_csv(filename, sep=None, header=True):
    r"""Read a CSV file.

    Parameters
    ----------
    filename : str
        Path to a CSV file.
    sep : str
        Separator.

    Returns
    -------
    data : dask dataframes

    Examples
    --------
    .. doctest::

        >>> from limix.io import read_csv
        >>> from limix.io.examples import csv_file_example
        >>>
        >>> df = read_csv(csv_file_example())
        >>> print(df.compute()) #doctest: +NORMALIZE_WHITESPACE
           pheno   attr1 attr2 attr3
        0    sex  string    10     a
        1   size   float    -3     b
        2  force     int     f     c
    """
    from dask.dataframe import read_csv as _read_csv

    if sep is None:
        sep = _infer_separator(filename)

    header = 0 if header else None
    return _read_csv(filename, sep=sep, header=header)


def see(filepath):
    """Shows a human-friendly representation of a CSV file.

    Parameters
    ----------
    filepath : str
        CSV file path.

    Returns
    -------
    str
        CSV representation.
    """
    from pandas import read_csv
    sep = _infer_separator(filepath)
    print(read_csv(filepath, sep=sep).describe())


def _count(candidates, line):
    counter = {c: 0 for c in candidates}
    for i in line:
        if i in candidates:
            counter[i] += 1
    return counter


def _update(counter, c):
    for (k, v) in c.items():
        if counter[k] != v:
            del counter[k]


def _infer_separator(fn):
    nmax = 9

    with open(fn, 'r') as f:
        line = f.readline().strip()
        counter = _count(set(line), line)

        for i in range(nmax - 1):
            line = f.readline().strip()
            if len(line) == 0:
                break
            c = _count(set(counter.keys()), line)
            _update(counter, c)
            if len(counter) == 1:
                return counter.keys()[0]

    for c in set([',', '\t', ' ']):
        if c in counter:
            return c

    counter = list(counter.items())
    counter = sorted(counter, key=lambda kv: kv[1])
    return counter[-1][0]



from numpy import ascontiguousarray, copyto, zeros
from tqdm import tqdm


def linear_kinship(G, out=None, verbose=True):
    r"""Estimate Kinship matrix via linear kernel.

    Examples
    --------
    .. doctest::

        >>> from numpy.random import RandomState
        >>> from limix.stats import linear_kinship
        >>>
        >>> random = RandomState(1)
        >>> X = random.randn(4, 100)
        >>> K = linear_kinship(X, verbose=False)
        >>> print(K)
        [[ 0.9131 -0.1928 -0.3413 -0.379 ]
         [-0.1928  0.8989 -0.2356 -0.4704]
         [-0.3413 -0.2356  0.9578 -0.3808]
         [-0.379  -0.4704 -0.3808  1.2302]]
    """
    (n, p) = G.shape
    if out is None:
        out = zeros((n, n))

    nsteps = min(100, p)

    for i in tqdm(range(nsteps), disable=not verbose):
        start = i * (p // nsteps)
        stop = min(start + p // nsteps, p)

        X = G[:, start:stop]
        X = X - X.mean(0)
        X /= X.std(0)

        out += ascontiguousarray(X.dot(X.T), float)

    out /= p

    return out


def gower_norm(K, out=None):
    r"""Perform Gower rescaling of covariance matrix K.

    The rescaled covariance matrix has sample variance of 1.

    Examples
    --------

    .. doctest::

        >>> from numpy.random import RandomState
        >>> from limix.stats import gower_norm
        >>> import scipy as sp
        >>>
        >>> random = RandomState(1)
        >>> X = random.randn(4, 4)
        >>> K = sp.dot(X,X.T)
        >>> Z = random.multivariate_normal(sp.zeros(4), K, 50)
        >>> print("%.3f" % sp.mean(Z.var(1,ddof=1)))
        2.335
        >>> Kn = gower_norm(K)
        >>> Zn = random.multivariate_normal(sp.zeros(4), Kn, 50)
        >>> print("%.3f" % sp.mean(Zn.var(1, ddof=1)))
        0.972
    """

    c = (K.shape[0] - 1) / (K.trace() - K.mean(0).sum())
    if out is None:
        return c * K

    copyto(out, K)
    out *= c



def query_and(*queries):
    """
    Given multiple queries it joins them using the & operator.

    Examples
    --------
    .. doctest::
        >>> from limix.data import query_and
        >>>
        >>> query1 = 'a < b'
        >>> query2 = 'b >= c'
        >>> query3 = 'c == 1'
        >>> query = query_and(query1, query2, query3)
        >>>
        >>> print(query)
        a < b & b >= c & c == 1
    """
    if len(queries) >= 1:
        query = ' & '.join(queries)
    else:
        query = None
    return query


def build_geno_query(idx_start=None,
                     idx_end=None,
                     pos_start=None,
                     pos_end=None,
                     chrom=None):
    """
    helper function to build genotype queries.

    Parameters
    ----------
    idx_start : int, optional
        start idx.
        If not None (default),
        the query 'idx >= idx_start' is considered.
    idx_end : int, optional
        end idx.
        If not None (default),
        the query 'idx < idx_end' is considered.
    pos_start : int, optional
        start chromosomal position.
        If not None (default),
        the query 'pos >= pos_start' is considered.
    pos_end : int, optional
        end chromosomal position.
        If not None (default),
        the query 'pos < pos_end' is considered.
    chrom : int, optional
        chromosome.
        If not None (default),
        the query 'chrom == chrom' is considered.

    Returns
    -------
        query : str

    Examples
    --------
    .. doctest::
        >>> from limix.data import build_geno_query
        >>>
        >>> query = build_geno_query(idx_start=4,
        ...                          idx_end=10,
        ...                          pos_start=45200,
        ...                          pos_end=80000,
        ...                          chrom=1)
        >>>
        >>> print(query)
        i >= 4 & i < 10 & pos >= 45200 & pos < 80000 & chrom == '1'
    """
    queries = []

    # gather all queries
    if idx_start is not None:
        query = "i >= %d" % idx_start
        queries.append(query)

    if idx_end is not None:
        query = "i < %d" % idx_end
        queries.append(query)

    if pos_start is not None:
        query = "pos >= %d" % pos_start
        queries.append(query)

    if pos_end is not None:
        query = "pos <% d" % pos_end
        queries.append(query)

    if chrom is not None:
        query = "chrom == '%s'" % str(chrom)
        queries.append(query)

    return query_and(*queries)



import asciitree
import h5py


class h5data_fetcher(object):
    r"""
    Fetch datasets from HDF5 files.

    Parameters
    ----------
    filename : str
        Filename to an HDF5 file.

    Examples
    --------
    .. doctest::

        >>> from limix.io import h5data_fetcher
        >>> from limix.io.examples import hdf5_file_example
        >>> with h5data_fetcher(hdf5_file_example()) as df:
        ...     X = df.fetch('/group/dataset')
        ...     print('%.4f' % X[0, 0].compute())
        -0.0453
    """

    def __init__(self, filename):
        self._filename = filename

    def __enter__(self):
        self._f = h5py.File(self._filename, 'r')
        return self

    def fetch(self, data_path):
        r"""
        Fetch a HDF5 dataset.

        Parameters
        ----------
        data_path : str
            Path to a dataset.

        Returns
        -------
        X : dask array
        """
        from dask.array import from_array
        data = self._f[data_path]
        if data.chunks is None:
            chunks = data.shape
        else:
            chunks = data.chunks
        return from_array(data, chunks=chunks)

    def __exit__(self, *exc):
        self._f.close()


def fetch(fp, path):
    """Fetches an array from hdf5 file.
    :param str fp: hdf5 file path.
    :param str path: path inside the hdf5 file.
    :returns: An :class:`numpy.ndarray` representation of the corresponding hdf5 dataset.
    """
    with h5py.File(fp, 'r') as f:
        return f[path][:]


def see(f_or_filepath, root_name='/', ret=False, show_chunks=False):
    """Shows a human-friendly tree representation of the contents of
    a hdf5 file.
    :param f_or_filepath: hdf5 file path or a reference to an open one.
    :param str root_name: group to be the root of the tree.
    :param bool ret: Whether to return a string or print it.
    :param bool show_chunks: show the chunks.
    :returns str: String representation if is `ret=True`.
    """
    if isinstance(f_or_filepath, str):
        with h5py.File(f_or_filepath, 'r') as f:
            return _tree(f, root_name, ret, show_chunks)
    else:
        return _tree(f_or_filepath, root_name, ret, show_chunks)


def _findnth(haystack, needle, n):
    parts = haystack.split(needle, n + 1)
    if len(parts) <= n + 1:
        return -1
    return len(haystack) - len(parts[-1]) - len(needle)


def _visititems(root, func, level=0, prefix=''):
    if root.name != '/':
        name = root.name
        eman = name[::-1]
        i1 = _findnth(eman, '/', level)
        name = '/' + eman[:i1][::-1]
        func(prefix + name, root)
    if not hasattr(root, 'keys'):
        return
    for k in root.keys():
        if root.file == root[k].file:
            _visititems(root[k], func, level + 1, prefix)
        else:
            _visititems(root[k], func, 0, prefix + root.name)


def _tree(f, root_name='/', ret=False, show_chunks=False):
    _names = []

    def get_names(name, obj):
        if isinstance(obj, h5py.Dataset):
            dtype = str(obj.dtype)
            shape = str(obj.shape)
            if show_chunks:
                chunks = str(obj.chunks)
                _names.append("%s [%s, %s, %s]" % (name[1:], dtype, shape,
                                                   chunks))
            else:
                _names.append("%s [%s, %s]" % (name[1:], dtype, shape))
        else:
            _names.append(name[1:])

    _visititems(f, get_names)

    class Node(object):
        def __init__(self, name, children):
            self.name = name
            self.children = children

        def __str__(self):
            return self.name

    root = Node(root_name, dict())

    def add_to_node(node, ns):
        if len(ns) == 0:
            return
        if ns[0] not in node.children:
            node.children[ns[0]] = Node(ns[0], dict())
        add_to_node(node.children[ns[0]], ns[1:])

    _names = sorted(_names)
    for n in _names:
        ns = n.split('/')
        add_to_node(root, ns)

    def child_iter(node):
        from numpy import argsort, asarray

        keys = list(node.children.keys())
        indices = argsort(keys)
        indices = asarray(indices)
        vals = list(node.children.values())
        return list(asarray(vals)[indices])

    msg = asciitree.draw_tree(root, child_iter)
    if ret:
        return msg
    print(msg)



import sys
import tempfile

PY2 = sys.version_info < (3, )

if PY2:
    import shutil
    from contextlib import contextmanager

    @contextmanager
    class TemporaryDirectory(object):
        def __init__(self):
            self._filepath = None

        def __enter__(self):
            self._filepath = tempfile.mkdtemp()
            return self._filepath

        def __exit__(self, *args):
            if self._filepath is not None:
                shutil.rmtree(self._filepath)
else:
    from tempfile import TemporaryDirectory



from optparse import OptionParser


def entry_point():
    from ..mtset.core.simPhenoCore import simPheno

    parser = OptionParser()
    parser.add_option("--bfile", dest='bfile', type=str, default=None)
    parser.add_option("--cfile", dest='cfile', type=str, default=None)
    parser.add_option("--pfile", dest='pfile', type=str, default=None)

    parser.add_option("--seed", dest='seed', type=int, default=0)
    parser.add_option("--nTraits", dest='nTraits', type=int, default=4)
    parser.add_option("--windowSize", dest='windowSize', type=int, default=3e4)

    parser.add_option("--chrom", dest='chrom', type=int, default=None)
    parser.add_option("--minPos", dest='pos_min', type=int, default=None)
    parser.add_option("--maxPos", dest='pos_max', type=int, default=None)

    parser.add_option("--vTotR", dest='vTotR', type=float, default=0.05)
    parser.add_option("--nCausalR", dest='nCausalR', type=int, default=8)
    parser.add_option("--pCommonR", dest='pCommonR', type=float, default=0.5)
    parser.add_option("--vTotBg", dest='vTotBg', type=float, default=0.4)
    parser.add_option("--pHidden", dest='pHidden', type=float, default=0.6)
    parser.add_option("--pCommon", dest='pCommon', type=float, default=0.5)
    (options, args) = parser.parse_args()

    simPheno(options)



import pandas as pd
import scipy as sp


def estCumPos(position, offset=0, chrom_len=None, return_chromstart=False):
    r"""
    Compute the cumulative position of variants from position dataframe

    Args:
        position (list or :class:`pandas.DataFrame`):
            positions in chromosome/chromosomal basepair position format.
            It can be specified as

            - list ``[chrom, pos]`` where ``chrom`` and ``pos`` are
              *ndarray* with chromosome values and basepair positions;
            - pandas DataFrame of chromosome values (key='chrom') and
              basepair positions (key='pos').

        chrom_len (ndarray, optional):
            vector with predefined chromosome length.
            By default, the length of the chromosome is taken to be the
            maximum basepair position (key='pos') in ``position`` on that
            chromosome.
        offset (float, optional):
            offset between chromosomes for cumulative position
            (default is 0 bp).
        return_chromstart (bool, optional):
            if ``True``, starting cumulative position of each chromosome
            is also returned (default is ``False``).

    Returns:
        (tuple): tuple containing:
            - **pos_cum** (*ndarray*):
              cumulative positions.
            - **chromstart** (*array_like*):
              starting cumulative positions for each chromosome.
              Returned only if ``return_chromstart=True``.

    Examples
    --------

        This function can be applied on a list of chrom and pos arrays

        .. doctest::

            >>> import scipy as sp
            >>> import pandas as pd
            >>> from limix.util import estCumPos
            >>>
            >>> pos = sp.kron(sp.ones(2), sp.arange(1,5)).astype(int)
            >>> chrom = sp.kron(sp.arange(1,3), sp.ones(4)).astype(int)
            >>>
            >>> pos_cum, chromstart = estCumPos([chrom, pos],
            ...                                 return_chromstart=True)
            >>>
            >>> print(chrom) # doctest: +SKIP
            [1 1 1 1 2 2 2 2]
            >>>
            >>> print(pos) # doctest: +SKIP
            [1 2 3 4 1 2 3 4]
            >>>
            >>> print(pos_cum) # doctest: +SKIP
            [1 2 3 4 5 6 7 8]
            >>>
            >>> print(chromstart) # doctest: +SKIP
            [1 5]

        or on a position dataframe:

        .. doctest::

            >>> position = pd.DataFrame(sp.array([chrom, pos]).T,
            ...                         columns=['chrom', 'pos'])
            >>> pos_cum, chromstart = estCumPos(position,
            ...                                 return_chromstart=True)
            >>> position['pos_cum'] = pos_cum
            >>> print(position)
               chrom  pos  pos_cum
            0      1    1        1
            1      1    2        2
            2      1    3        3
            3      1    4        4
            4      2    1        5
            5      2    2        6
            6      2    3        7
            7      2    4        8
    """
    if type(position) != pd.core.frame.DataFrame:

        if type(position) != list:
            raise TypeError('position must be a dataframe or list of arrays')

        chrom = position[0]
        pos = position[1]

        if type(pos) != sp.ndarray or type(chrom) != sp.ndarray:
            raise TypeError('position must be a dataframe or list of arrays')

        if len(pos.shape) != 1 or len(chrom.shape) != 1:
            raise ValueError('pos and chrom should be arrays')

        if len(pos) != len(chrom):
            raise ValueError('pos and chrom should have the same length')

        position = pd.DataFrame(
            sp.array([pos, chrom]).T, columns=['pos', 'chrom'])

    RV = position.copy()

    chromvals = sp.unique(position['chrom'])  # sp.unique is always sorted
    chromstart = sp.zeros_like(chromvals)
    pos_cum = sp.zeros_like(position.shape[0])
    if 'pos_cum' not in position:
        RV["pos_cum"] = sp.zeros_like(position['pos'])
    pos_cum = RV['pos_cum'].values
    to_add = 0
    for i, mychrom in enumerate(chromvals):
        i_chr = position['chrom'] == mychrom
        if chrom_len is None:
            maxpos = position['pos'][i_chr].max() + offset
        else:
            maxpos = chrom_len[i] + offset
        pos_cum[i_chr.values] = to_add + position.loc[i_chr, 'pos']
        chromstart[i] = pos_cum[i_chr.values].min()
        to_add += maxpos

    if return_chromstart:
        return pos_cum, chromstart
    else:
        return pos_cum


def unique_variants(snps, return_idxs=False):
    r"""
    Filters out variants with the same genetic profile.

    Args:
        snps (ndarray):
            (`N`, `S`) ndarray of genotype values for `N` individuals and
            `S` variants.

    Returns:
        ndarray: genotype array with unique variants.

    Examples
    --------

        .. doctest::

            >>> from numpy.random import RandomState
            >>> from numpy import kron, ones
            >>> from limix.util import unique_variants
            >>> from numpy import set_printoptions
            >>> set_printoptions(4)
            >>> random = RandomState(1)
            >>>
            >>> N = 4
            >>> snps = kron(random.randn(N,3)<0., ones((1,2)))
            >>>
            >>> print(snps) # doctest: +SKIP
            [[0. 0. 1. 1. 1. 1.]
             [1. 1. 0. 0. 1. 1.]
             [0. 0. 1. 1. 0. 0.]
             [1. 1. 0. 0. 1. 1.]]
            >>>
            >>> snps_u = unique_variants(snps)
            >>>
            >>> print(snps_u) # doctest: +SKIP
            [[0. 1. 1.]
             [1. 0. 1.]
             [0. 1. 0.]
             [1. 0. 1.]]
    """

    _s = sp.dot(sp.rand(snps.shape[0]), snps)

    v, ix = sp.unique(_s, return_index=True)
    idxs_u = sp.sort(ix)

    snps = snps[:, idxs_u]

    if return_idxs:
        return (snps, ix)

    return snps



import os
import time as TIME
import warnings

import h5py
import scipy as sp
import scipy.linalg as la


class MTSet():
    r"""
    Multi-trait set tests (mtSet).

    Mixed-model approach that enables joint analysis across multiple correlated
    traits while accounting for population structure and relatedness [CRS15]_.

    Parameters
    ----------
    Y : (`N`, `P`) ndarray
        Individuals by phenotypes.
    F : (`N`, `K`) ndarray, optional
        Individuals by covariates. By default, no fixed effect term is set.
    R : (`N`, `N`) ndarray, optional
        LMM-covariance/kinship coefficients.
        ``U_R`` and ``S_R`` can be provided instead of ``R``.
        If neither ``R`` nor ``U_R`` and ``S_R`` are provided,
        the null models has iid normal residuals.
    U_R : (`N`, `N`) ndarray, optional
        Eigenvectors of ``R``.
        ``U_R`` and ``S_R`` can be provided instead of ``R``.
        If neither ``R`` nor ``U_R`` and ``S_R`` are provided,
        iid normal residuals are considered.
    S_R : (`N`, ) ndarray, optional
        Eigenvalues of ``R``.
        ``U_R`` and ``S_R`` can be provided instead of ``R``.
        If neither ``R`` nor ``U_R`` and ``S_R`` are provided,
        iid normal residuals are considered.
    traitID : ndarray, optional
        Phenotype IDs.
    rank : int, optional
        Rank of the trait set covariance. The default value is 1.

    Examples
    --------
    This example shows how to fit mtSet when modelling population
    structure/relatedness using the full genetic relatedness matrix.

    .. doctest::

        >>> from numpy.random import RandomState
        >>> from limix.mtset import MTSet
        >>> from numpy import dot, eye, ones, concatenate
        >>> from numpy import set_printoptions
        >>> set_printoptions(4)
        >>>
        >>> random = RandomState(1)
        >>>
        >>> N = 100
        >>> P = 2
        >>>
        >>> pheno = random.randn(N, P)
        >>> pheno-= pheno.mean(0)
        >>> pheno/= pheno.std(0)
        >>>
        >>> W = random.randn(N, 10)
        >>> kinship = dot(W, W.T) / float(10)
        >>> kinship+= 1e-4 * eye(N)
        >>>
        >>> mtset = MTSet(pheno, R=kinship)
        >>> res_null = mtset.fitNull()
        >>>
        >>> print(res_null['conv'][0])
        True
        >>> print('%.2f'%res_null['NLL0'])
        98.43
        >>>
        >>> S = 4
        >>> snp_set = (random.rand(N, S) < 0.2).astype(float)
        >>> res = mtset.optimize(snp_set)
        >>>
        >>> print("%.4f" % res['LLR'][0])
        0.0616
        >>> print(res['Cr']) # doctest: +SKIP
        [[0.0006 0.0025]
         [0.0025 0.0113]]

    This example shows how to fit mtSet when modelling population
    structure/relatedness by introducing the top principle components
    of the genetic relatedness matrix (``pc_rrm``) as fixed effects.

        >>> random = RandomState(0)
        >>>
        >>> mean = ones((N, 1))
        >>> pc_rrm = random.randn(N, 4)
        >>> covs = concatenate([mean, pc_rrm], 1)
        >>>
        >>> mtset = MTSet(pheno, F=covs)
        >>> res_null = mtset.fitNull()
        >>>
        >>> print(res_null['conv'][0])
        True
        >>> print('%.2f'%res_null['NLL0'])
        118.36
        >>>
        >>> res = mtset.optimize(snp_set)
        >>>
        >>> print("%.4f" % res['LLR'][0])
        0.1373
        >>> print(res['Cr']) # doctest: +SKIP
        [[0.0002 0.0019]
         [0.0019 0.0207]]
    """

    def __init__(self,
                 Y=None,
                 R=None,
                 S_R=None,
                 U_R=None,
                 traitID=None,
                 F=None,
                 rank=1):
        from limix_core.gp import GP2KronSum
        from limix_core.gp import GP2KronSumLR
        from limix_core.gp import GP3KronSumLR
        from limix_core.covar import FreeFormCov
        # data
        noneNone = S_R is not None and U_R is not None
        self.bgRE = R is not None or noneNone
        # fixed effect
        msg = 'The current implementation of the full rank mtSet'
        msg += ' does not support covariates.'
        msg += ' We reccommend to regress out covariates and'
        msg += ' subsequently quantile normalize the phenotypes'
        msg += ' to a normal distribution prior to use mtSet.'
        msg += ' This can be done within the LIMIX framework using'
        msg += ' the methods limix.util.preprocess.regressOut and'
        msg += ' limix.util.preprocess.gaussianize'
        assert not (F is not None and self.bgRE), msg
        from limix.util.preprocess import remove_dependent_cols
        if F is not None:
            F = remove_dependent_cols(F)
            A = sp.eye(Y.shape[1])
        else:
            A = None
        # traitID
        if traitID is None:
            traitID = sp.array(['trait %d' % p for p in range(Y.shape[1])])
        self.setTraitID(traitID)
        # init covariance matrices and gp
        Cg = FreeFormCov(Y.shape[1])
        Cn = FreeFormCov(Y.shape[1])
        G = 1. * (sp.rand(Y.shape[0], 1) < 0.2)
        if self.bgRE:
            self._gp = GP3KronSumLR(
                Y=Y, Cg=Cg, Cn=Cn, R=R, S_R=S_R, U_R=U_R, G=G, rank=rank)
        else:
            self._gp = GP2KronSumLR(Y=Y, Cn=Cn, G=G, F=F, A=A)
        # null model params
        self.null = None
        # calls itself for column-by-column trait analysis
        self.stSet = None
        self.nullST = None
        self.infoOpt = None
        self.infoOptST = None
        pass

    ##################################################
    # Properties
    ##################################################
    @property
    def N(self):
        return self._gp.covar.dim_r

    @property
    def P(self):
        return self._gp.covar.dim_c

    @property
    def Cr(self):
        return self._gp.covar.Cr

    @property
    def Cg(self):
        return self._gp.covar.Cg

    @property
    def Cn(self):
        return self._gp.covar.Cn

    @property
    def Y(self):
        return self._gp.mean.Y

    @property
    def F(self):
        try:
            return self._gp.mean.F[0]
        except BaseException:
            return None

    @property
    def A(self):
        try:
            return self._gp.mean.A[0]
        except BaseException:
            return None

    @property
    def S_R(self):
        if self.bgRE:
            RV = self._gp.covar.Sr()
        else:
            RV = None
        return RV

    @property
    def U_R(self):
        if self.bgRE:
            RV = self._gp.covar.Lr().T
        else:
            RV = None
        return RV

    #########################################
    # Setters
    #########################################
    @Y.setter
    def Y(self, value):
        assert value.shape[0] == self.N, 'Dimension mismatch'
        assert value.shape[1] == self.P, 'Dimension mismatch'
        self._gp.mean.Y = Y

    def setTraitID(self, traitID):
        self.traitID = traitID

    ################################################
    # Fitting null model
    ###############################################
    def fitNull(self,
                cache=False,
                out_dir='./cache',
                fname=None,
                rewrite=False,
                seed=None,
                factr=1e3,
                n_times=10,
                init_method=None,
                verbose=False):
        r"""
        Fit null model

        Args:
            verbose ()
            cache (bool, optional):
                If False (default), the null model is fitted and
                the results are not cached.
                If True, the cache is activated.
                The cache file dir and name can be specified using
                ``hcache`` and ``fname``.
                When ``cache=True``, we distinguish the following cases:

                - if the specified file does not exist,
                  the output of the null model fiting is cached in the file.
                - if the specified file exists and ``rewrite=True``,
                  the cache file is overwritten.
                - if the specified file exists and ``rewrite=False``,
                  the results from the cache file are imported
                  (the null model is not re-fitted).

            out_dir (str, optional):
                output dir of the cache file.
                The default value is "./cache".
            fname (str, optional):
                Name of the cache hdf5 file.
                It must be specified if ``cache=True``.
            rewrite (bool, optional):
                It has effect only if cache `cache=True``.
                In this case, if ``True``,
                the cache file is overwritten in case it exists.
                The default value is ``False``
            factr (float, optional):
                optimization paramenter that determines the accuracy
                of the solution.
                By default it is 1000.
                (see scipy.optimize.fmin_l_bfgs_b for more details).
            verbose (bool, optional):
                verbose flag.

        Returns:
            (dict): dictionary containing:
                - **B** (*ndarray*): estimated effect sizes (null);
                - **Cg** (*ndarray*): estimated relatedness trait covariance (null);
                - **Cn** (*ndarray*): estimated genetic noise covariance (null);
                - **conv** (*bool*): convergence indicator;
                - **NLL0** (*ndarray*): negative loglikelihood (NLL) of the null model;
                - **LMLgrad** (*ndarray*): norm of the gradient of the NLL.
                - **time** (*time*): elapsed time (in seconds).
        """
        from limix_core.gp import GP2KronSum
        from limix_core.gp import GP2KronSumLR
        from limix_core.gp import GP3KronSumLR
        from limix_core.covar import FreeFormCov
        if seed is not None:
            sp.random.seed(seed)

        read_from_file = False
        if cache:
            assert fname is not None, 'MultiTraitSetTest:: specify fname'
            if not os.path.exists(out_dir):
                os.makedirs(out_dir)
            out_file = os.path.join(out_dir, fname)
            read_from_file = os.path.exists(out_file) and not rewrite

        RV = {}
        if read_from_file:
            f = h5py.File(out_file, 'r')
            for key in list(f.keys()):
                RV[key] = f[key][:]
            f.close()
            self.setNull(RV)
        else:
            start = TIME.time()
            if self.bgRE:
                self._gpNull = GP2KronSum(
                    Y=self.Y,
                    F=None,
                    A=None,
                    Cg=self.Cg,
                    Cn=self.Cn,
                    R=None,
                    S_R=self.S_R,
                    U_R=self.U_R)
            else:
                self._gpNull = GP2KronSumLR(
                    self.Y,
                    self.Cn,
                    G=sp.ones((self.N, 1)),
                    F=self.F,
                    A=self.A)
                # freezes Cg to 0
                n_params = self._gpNull.covar.Cr.getNumberParams()
                self._gpNull.covar.Cr.setParams(1e-9 * sp.ones(n_params))
                self._gpNull.covar.act_Cr = False
            for i in range(n_times):
                params0 = self._initParams(init_method=init_method)
                self._gpNull.setParams(params0)
                conv, info = self._gpNull.optimize(
                    verbose=verbose, factr=factr)
                if conv:
                    break
            if not conv:
                warnings.warn("not converged")
            LMLgrad = (self._gpNull.LML_grad()['covar']**2).mean()
            LML = self._gpNull.LML()
            if self._gpNull.mean.n_terms == 1:
                RV['B'] = self._gpNull.mean.B[0]
            elif self._gpNull.mean.n_terms > 1:
                warning.warn('generalize to more than 1 fixed effect term')
            if self.bgRE:
                RV['params0_g'] = self.Cg.getParams()
            else:
                RV['params0_g'] = sp.zeros_like(self.Cn.getParams())
            RV['params0_n'] = self.Cn.getParams()
            if self.bgRE:
                RV['Cg'] = self.Cg.K()
            else:
                RV['Cg'] = sp.zeros_like(self.Cn.K())
            RV['Cn'] = self.Cn.K()
            RV['conv'] = sp.array([conv])
            RV['time'] = sp.array([TIME.time() - start])
            RV['NLL0'] = sp.array([LML])
            RV['LMLgrad'] = sp.array([LMLgrad])
            RV['nit'] = sp.array([info['nit']])
            RV['funcalls'] = sp.array([info['funcalls']])
            self.null = RV
            from limix.util.util_functions import smartDumpDictHdf5
            if cache:
                f = h5py.File(out_file, 'w')
                smartDumpDictHdf5(RV, f)
                f.close()
        return RV

    def getNull(self):
        """ get null model info """
        return self.null

    def setNull(self, null):
        """ set null model info """
        self.null = null

    ###########################################
    # Fitting alternative model
    ###########################################

    def optimize(self,
                 G,
                 params0=None,
                 n_times=10,
                 vmax=5,
                 factr=1e7,
                 verbose=False):
        """
        Fit alternative model

        Args:
            G (ndarray):
                (N, S) genotype values for `N` samples and `S` variants
                (defines the set component)
            params0 (ndarray, optional):
                initial parameter values for optimization.
                By default the parameters are initialized
                by randonmly pertumrning the maximum likelihood
                estimators (MLE) under the null.
            n_times (ndarray, optional):
                maximum number of restarts in case the optimization does not
                converge.
                It has no effect if ``params0`` is provided.
            vmax (float, option):
                maximum value for variance parameters that is accepted.
                If any of the variance MLE is ``>vmax`` then we say that
                the optimization has not converged.
                Default value is 5.
            factr (float, optional):
                optimization paramenter that determines the accuracy
                of the solution.
                By default it is 1e7.
                (see scipy.optimize.fmin_l_bfgs_b for more details).
            verbose (bool, optional):
                verbose flag.

        Returns:
            (dict): dictionary containing:
                - **Cr** (*ndarray*): estimated set trait covariance (alt);
                - **Cg** (*ndarray*): estimated relatedness trait covariance (alt);
                - **Cn** (*ndarray*): estimated genetic noise covariance (alt);
                - **conv** (*ndarray*): convergence indicator;
                - **NLLAlt** (*ndarray*): negative loglikelihood (NLL) of the alternative model;
                - **LMLgrad** (*ndarray*): norm of the gradient of the NLL.
                - **LLR** (*ndarray**): log likelihood ratio statistics;
                - **var** (*ndarray*): (`P`, `K`) ndarray of  variance estimates
                  for the ``P`` traits and the ``K``
                - **time** (*ndarray*): elapsed time (in seconds).
                - **nit** (*ndarray*):
                  number of iterations L-BFGS.
                  (see scipy.optimize.fmin_l_bfgs_b for more details).
        """
        # set params0 from null if params0 is None
        if params0 is None:
            if self.null is None:
                if verbose:
                    print(".. fitting null model")
                self.fitNull()
            if self.bgRE:
                params0 = sp.concatenate(
                    [self.null['params0_g'], self.null['params0_n']])
            else:
                params0 = self.null['params0_n']
            params_was_None = True
        else:
            params_was_None = False
        G *= sp.sqrt(self.N / (G**2).sum())
        self._gp.covar.G = G
        start = TIME.time()
        for i in range(n_times):
            if params_was_None:
                n_params = self.Cr.getNumberParams()
                _params0 = {
                    'covar':
                    sp.concatenate([1e-3 * sp.randn(n_params), params0])
                }
            else:
                _params0 = {'covar': params0}
            self._gp.setParams(_params0)
            conv, info = self._gp.optimize(factr=factr, verbose=verbose)
            conv *= self.Cr.K().diagonal().max() < vmax
            conv *= self.getLMLgrad() < 0.1
            if conv or not params_was_None:
                break
        self.infoOpt = info
        if not conv:
            warnings.warn("not converged")
        # return value
        RV = {}
        if self.P > 1:
            RV['Cr'] = self.Cr.K()
            if self.bgRE:
                RV['Cg'] = self.Cg.K()
            RV['Cn'] = self.Cn.K()
        RV['time'] = sp.array([TIME.time() - start])
        RV['params0'] = _params0
        RV['nit'] = sp.array([info['nit']])
        RV['funcalls'] = sp.array([info['funcalls']])
        RV['var'] = self.getVariances()
        RV['conv'] = sp.array([conv])
        RV['NLLAlt'] = sp.array([self.getNLLAlt()])
        RV['LLR'] = sp.array([self.getLLR()])
        RV['LMLgrad'] = sp.array([self.getLMLgrad()])
        return RV

    def getInfoOpt(self):
        """ get information for the optimization """
        return self.infoOpt

    def getVariances(self):
        """
        get variances
        """
        var = []
        var.append(self.Cr.K().diagonal())
        if self.bgRE:
            var.append(self.Cg.K().diagonal())
        var.append(self.Cn.K().diagonal())
        var = sp.array(var)
        return var

    def getNLLAlt(self):
        """
        get negative log likelihood of the alternative
        """
        return self._gp.LML()

    def getLLR(self):
        """
        get log likelihood ratio
        """
        assert self.null is not None, 'null model needs to be fitted!'
        return self.null['NLL0'][0] - self.getNLLAlt()

    def getLMLgrad(self):
        """
        get norm LML gradient
        """
        return (self._gp.LML_grad()['covar']**2).mean()

    def fitNullTraitByTrait(self,
                            verbose=False,
                            cache=False,
                            out_dir='./cache',
                            fname=None,
                            rewrite=False):
        """
        Fit null model trait by trait
        """
        read_from_file = False
        if cache:
            assert fname is not None, 'MultiTraitSetTest:: specify fname'
            if not os.path.exists(out_dir):
                os.makedirs(out_dir)
            out_file = os.path.join(out_dir, fname)
            read_from_file = os.path.exists(out_file) and not rewrite

        RV = {}
        if read_from_file:
            f = h5py.File(out_file, 'r')
            for p in range(self.P):
                trait_id = self.traitID[p]
                g = f[trait_id]
                RV[trait_id] = {}
                for key in list(g.keys()):
                    RV[trait_id][key] = g[key][:]
            f.close()
            self.nullST = RV
        else:
            """ create stSet and fit null column by column returns all info """
            if self.stSet is None:
                y = sp.zeros((self.N, 1))
                self.stSet = MTSet(Y=y, S_R=self.S_R, U_R=self.U_R, F=self.F)
            RV = {}
            for p in range(self.P):
                trait_id = self.traitID[p]
                self.stSet.Y = self.Y[:, p:p + 1]
                RV[trait_id] = self.stSet.fitNull()
            self.nullST = RV
            from limix.util.util_functions import smartDumpDictHdf5
            if cache:
                f = h5py.File(out_file, 'w')
                smartDumpDictHdf5(RV, f)
                f.close()
        return RV

    def optimizeTraitByTrait(self, G, verbose=False, n_times=10, factr=1e3):
        """ Optimize trait by trait """
        assert self.nullST is not None, 'fit null model beforehand'
        RV = {}
        self.infoOptST = {}
        for p in range(self.P):
            trait_id = self.traitID[p]
            self.stSet.Y = self.Y[:, p:p + 1]
            self.stSet.setNull(self.nullST[trait_id])
            RV[trait_id] = self.stSet.optimize(
                G, n_times=n_times, factr=factr, verbose=verbose)
            self.infoOptST[trait_id] = self.stSet.getInfoOpt()
        return RV

    def getInfoOptST(self):
        """ get information on the optimization """
        return self.infoOptST

    def _initParams(self, init_method=None):
        """ internal function for params initialization """
        if self.bgRE:
            if init_method == 'random':
                params0 = {
                    'covar': sp.randn(self._gpNull.covar.getNumberParams())
                }
            else:
                if self.P == 1:
                    params0 = {'covar': sp.sqrt(0.5) * sp.ones(2)}
                else:
                    cov = 0.5 * sp.cov(self.Y.T) + 1e-4 * sp.eye(self.P)
                    chol = la.cholesky(cov, lower=True)
                    params = chol[sp.tril_indices(self.P)]
                    params0 = {'covar': sp.concatenate([params, params])}
        else:
            if self.P == 1:
                params_cn = sp.array([1.])
            else:
                cov = sp.cov(self.Y.T) + 1e-4 * sp.eye(self.P)
                chol = la.cholesky(cov, lower=True)
                params_cn = chol[sp.tril_indices(self.P)]
            params0 = {'covar': params_cn}
        return params0


if __name__ == '__main__':
    from limix.util.preprocess import covar_rescale

    sp.random.seed(0)

    # generate phenotype
    n = 1000
    p = 4
    f = 10
    Y = sp.randn(n, p)
    X = sp.randn(n, f)
    G = sp.randn(n, f)
    R = sp.dot(X, X.T)
    R = covar_rescale(R)

    mts = mtset(Y, R=R)
    nullMTInfo = mts.fitNull(cache=False)
    mts.optimize(G)



from __future__ import division

from numpy import ascontiguousarray, copy, ones, var
from numpy_sugar.linalg import economic_qs

from glimix_core.glmm import GLMM


def estimate(pheno, lik, K, covs=None, verbose=True):
    r"""Estimate the so-called narrow-sense heritability.

    It supports Normal, Bernoulli, Binomial, and Poisson phenotypes.
    Let :math:`N` be the sample size and :math:`S` the number of covariates.

    Parameters
    ----------
    pheno : tuple, array_like
        Phenotype. Dimensions :math:`N\\times 0`.
    lik : {'normal', 'bernoulli', 'binomial', 'poisson'}
        Likelihood name.
    K : array_like
        Kinship matrix. Dimensions :math:`N\\times N`.
    covs : array_like
        Covariates. Default is an offset. Dimensions :math:`N\\times S`.

    Returns
    -------
    float
        Estimated heritability.

    Examples
    --------
    .. doctest::

        >>> from numpy import dot, exp, sqrt
        >>> from numpy.random import RandomState
        >>> from limix.heritability import estimate
        >>>
        >>> random = RandomState(0)
        >>>
        >>> G = random.randn(50, 100)
        >>> K = dot(G, G.T)
        >>> z = dot(G, random.randn(100)) / sqrt(100)
        >>> y = random.poisson(exp(z))
        >>>
        >>> print('%.2f' % estimate(y, 'poisson', K, verbose=False))
        0.70
    """

    K = _background_standardize(K)
    QS = economic_qs(K)

    lik = lik.lower()

    if lik == 'binomial':
        p = len(pheno[0])
    else:
        p = len(pheno)

    if covs is None:
        covs = ones((p, 1))

    glmm = GLMM(pheno, lik, covs, QS)
    glmm.feed().maximize(verbose=verbose)

    g = glmm.scale * (1 - glmm.delta)
    e = glmm.scale * glmm.delta
    h2 = g / (var(glmm.mean()) + g + e)

    return h2


def _background_standardize(K):
    from ..stats.kinship import gower_norm

    K = copy(K, 'C')
    K = ascontiguousarray(K, dtype=float)
    gower_norm(K, K)
    K /= K.diagonal()

    return K



def boxcox(X):
    r"""Gaussianize X using the Box-Cox transformation.

    Each phentoype is brought to a positive schale by first subtracting the
    minimum value and adding 1.
    Then each phenotype is transformed by the Box-Cox transformation.

    Args:
        X (array_like): samples by phenotypes.

    Returns:
        array_like: Box-Cox power transformed array.

    Examples
    --------

        .. doctest::

            >>> from numpy.random import RandomState
            >>> from limix.stats import boxcox
            >>> from numpy import set_printoptions
            >>> set_printoptions(4)
            >>>
            >>> random = RandomState(0)
            >>> X = random.randn(5, 2)
            >>>
            >>> print(boxcox(X)) # doctest: +SKIP
            [[2.7136 0.9544]
             [1.3844 1.6946]
             [2.9066 0.    ]
             [1.3407 0.644 ]
             [0.     0.9597]]
    """
    from limix.util.preprocess import boxcox as _boxcox

    return _boxcox(X)[0]



import pdb

import scipy as sp
import scipy.stats as st


class Chi2mixture(object):
    """
    A class for continuous random variable following a chi2 mixture

    Class for evaluation of P values for a test statistic that follows a
    two-component mixture of chi2

    .. math::

        (1-\pi)\chi^2(0) + \pi a \chi^2(d).

    Here :math:`\pi` is the probability being in the first component and
    :math:`a` and :math:`d` are the scale parameter and the number of
    degrees of freedom of the second component.

    Args:
        scale_min (float): minimum value used for fitting the scale parameter
        scale_max (float): maximum value used for fitting the scale parameter
        dofmin (float): minimum value used for fitting the dof parameter
        dofmax (float): maximum value used for fitting the dof parameter
        qmax (float): only the top qmax quantile is used for the fit
        n_interval (int): number of intervals when performing gridsearch
        tol (float): tolerance of being zero

    Examples
    --------

        .. doctest::

            >>> from numpy.random import RandomState
            >>> import scipy as sp
            >>> from limix.stats import Chi2mixture
            >>>
            >>> scale = 0.3
            >>> dof = 2
            >>> mixture = 0.2
            >>> n = 100
            >>>
            >>> random = RandomState(1)
            >>> x =  random.chisquare(dof, n)
            >>> n0 = int( (1-mixture) * n)
            >>> idxs = random.choice(n, n0, replace=False)
            >>> x[idxs] = 0
            >>>
            >>> chi2mix = Chi2mixture(scale_min=0.1, scale_max=5.0,
            ...                       dof_min=0.1, dof_max=5.0,
            ...                       qmax=0.1, tol=4e-3)
            >>> chi2mix.estimate_chi2mixture(x)
            >>> pv = chi2mix.sf(x)
            >>> print(pv[:4]) # doctest: +SKIP
            [0.2 0.2 0.2 0.2]
            >>>
            >>> print('%.2f' % chi2mix.scale)
            1.98
            >>> print('%.2f' % chi2mix.dof)
            0.89
            >>> print('%.2f' % chi2mix.mixture)
            0.20
    """

    def __init__(self,
                 scale_min=0.1,
                 scale_max=5.0,
                 dof_min=0.1,
                 dof_max=5.0,
                 n_intervals=100,
                 qmax=0.1,
                 tol=0):

        self.scale_min = scale_min
        self.scale_max = scale_max
        self.dof_min = dof_min
        self.dof_max = dof_max
        self.qmax = qmax
        self.n_intervals = n_intervals
        self.tol = tol

    def estimate_chi2mixture(self, lrt):
        """
        Estimates the parameters of the mixture of chi2 by fitting the
        empirical distribution of null test statistic.

        Args:
            lrt (array_like): null test statistcs.
        """

        # step 1: estimate the probability of being in component one
        self.mixture = 1 - (lrt <= self.tol).mean()
        n_false = sp.sum(lrt > self.tol)

        # step 2: only use the largest qmax fraction of test statistics to estimate the
        #           remaining parameters
        n_fitting = int(sp.ceil(self.qmax * n_false))
        lrt_sorted = -sp.sort(-lrt)[:n_fitting]
        q = sp.linspace(0, 1, n_false)[1:n_fitting + 1]
        log_q = sp.log10(q)

        # step 3: fitting scale and dof by minimizing the squared error of the log10 p-values
        #        with their theorietical values [uniform distribution]
        MSE_opt = sp.inf
        MSE = sp.zeros((self.n_intervals, self.n_intervals))

        for i, scale in enumerate(
                sp.linspace(self.scale_min, self.scale_max, self.n_intervals)):
            for j, dof in enumerate(
                    sp.linspace(self.dof_min, self.dof_max, self.n_intervals)):
                p = st.chi2.sf(lrt_sorted / scale, dof)
                log_p = sp.log10(p)
                MSE[i, j] = sp.mean((log_q - log_p)**2)
                if MSE[i, j] < MSE_opt:
                    MSE_opt = MSE[i, j]
                    self.scale = scale
                    self.dof = dof

    def sf(self, lrt):
        """
        Computes the P values from test statistics lrt

        Args:
            lrt (array_like): test statistics.

        Returns:
            array_like: pvalues
        """
        _lrt = sp.copy(lrt)
        _lrt[lrt < self.tol] = 0
        pv = self.mixture * st.chi2.sf(_lrt / self.scale, self.dof)
        return pv



from limix.data import build_geno_query


class GIter:
    def __init__(self, geno_reader, start=0, end=None, batch_size=1000):
        self.gr = geno_reader
        self.current = start
        if end is None:
            end = self.gr.getSnpInfo().shape[0]
        self.end = end
        self.batch_size = batch_size

    def __iter__(self):
        return self

    def __next__(self):
        if self.current >= self.end:
            raise StopIteration
        else:
            _end = self.current + self.batch_size
            query = build_geno_query(idx_start=self.current, idx_end=_end)
            rv = self.gr.subset_snps(query)
            self.current = _end
            return rv

    def next(self):  # for python2 compatibility
        return self.__next__()



import scipy as sp
import scipy.linalg as la
from .linalg_utils import lowrank_approx

ntype_dict = {'assoc': 'null', 'gxe': 'block', 'gxehet': 'rank1'}


def define_gp(Y, Xr, Sg, Ug, type):
    from limix_core.covar import LowRankCov
    from limix_core.covar import FixedCov
    from limix_core.covar import FreeFormCov
    from limix_core.gp import GP3KronSumLR
    from limix_core.gp import GP2KronSum
    P = Y.shape[1]
    _A = sp.eye(P)
    if type in 'rank1':
        _Cr = LowRankCov(P, 1)
    elif type == 'block':
        _Cr = FixedCov(sp.ones((P, P)))
    elif type == 'full':
        _Cr = FreeFormCov(P)
    elif type == 'null':
        pass
    else:
        print('poppo')
    _Cn = FreeFormCov(P)
    _Cg = FreeFormCov(P)
    if type == 'null':
        _gp = GP2KronSum(Y=Y, Cg=_Cg, Cn=_Cn, S_R=Sg, U_R=Ug)
    else:
        _gp = GP3KronSumLR(Y=Y, G=Xr, Cr=_Cr, Cg=_Cg, Cn=_Cn, S_R=Sg, U_R=Ug)
    return _gp


class MvSetTestFull():

    def __init__(
            self,
            Y=None,
            Xr=None,
            Rg=None,
            Ug=None,
            Sg=None,
            factr=1e7,
            debug=False):
        """
        Args:
            Y:          [N, P] phenotype matrix
            Xr:         [N, S] genotype data of the set component
            R:          [N, S] genotype data of the set component
            factr:      paramenter that determines the accuracy of the solution
                        (see scipy.optimize.fmin_l_bfgs_b for more details)
        """
        # assert Xr
        Xr -= Xr.mean(0)
        Xr /= Xr.std(0)
        Xr /= sp.sqrt(Xr.shape[1])
        self.Y = Y
        self.Xr = Xr
        if Sg is None or Ug is None:
            Sg, Ug = la.eigh(Rg)
        self.Rg = Rg
        self.Ug = Ug
        self.Sg = Sg
        self.covY = sp.cov(Y.T)
        self.factr = factr
        self.debug = debug
        self.gp = {}
        self.info = {}
        # _trRr = sp.diagonal(sp.dot(self.Ug, sp.dot(sp.diag(self.Sg), self.Ug.T))).sum()
        self.trRg = ((self.Ug * self.Sg**0.5)**2).sum()

    def assoc(self):
        # fit model
        for key in ['null', 'full']:
            if key not in list(self.gp.keys()):
                if self.debug:
                    print('.. dening %s' % key)
                self.gp[key] = define_gp(
                    self.Y, self.Xr, self.Sg, self.Ug, key)
                if self.debug:
                    print('.. fitting %s' % key)
                self.info[key] = self._fit(key, vc=True)
        return self.info['null']['LML'] - self.info['full']['LML']

    def gxe(self):
        # fit model
        for key in ['null', 'full', 'block']:
            if key not in list(self.gp.keys()):
                if self.debug:
                    print('.. defining %s' % key)
                self.gp[key] = define_gp(
                    self.Y, self.Xr, self.Sg, self.Ug, key)
                if self.debug:
                    print('.. fitting %s' % key)
                self.info[key] = self._fit(key, vc=True)
        return self.info['block']['LML'] - self.info['full']['LML']

    def gxehet(self):
        # fit model
        for key in ['null', 'full', 'rank1']:
            if key not in list(self.gp.keys()):
                if self.debug:
                    print('.. defining %s' % key)
                self.gp[key] = define_gp(
                    self.Y, self.Xr, self.Sg, self.Ug, key)
                if self.debug:
                    print('.. fitting %s' % key)
                self.info[key] = self._fit(key, vc=True)
        return self.info['rank1']['LML'] - self.info['full']['LML']

    def assoc_null(self, n_nulls=30):
        LLR0 = sp.zeros(n_nulls)
        for ni in range(n_nulls):
            idx_perms = sp.random.permutation(self.Y.shape[0])
            _Xr = self.Xr[idx_perms]
            mvset0 = MvSetTestFull(Y=self.Y, Sg=self.Sg, Ug=self.Ug, Xr=_Xr)
            LLR0[ni] = mvset0.assoc()
        return LLR0

    def gxe_null(self, n_nulls=30):
        LLR0 = sp.zeros(n_nulls)
        for ni in range(n_nulls):
            _Y = self.gp['block'].simulate_pheno()
            mvset0 = MvSetTestFull(Y=_Y, Sg=self.Sg, Ug=self.Ug, Xr=self.Xr)
            LLR0[ni] = mvset0.gxe()
        return LLR0

    def gxehet_null(self, n_nulls=30):
        LLR0 = sp.zeros(n_nulls)
        for ni in range(n_nulls):
            _Y = self.gp['rank1'].simulate_pheno()
            mvset0 = MvSetTestFull(Y=_Y, Sg=self.Sg, Ug=self.Ug, Xr=self.Xr)
            LLR0[ni] = mvset0.gxehet()
        return LLR0

    def score(self):
        pass

    def _fit(self, type, vc=False):
        # 2. init
        if type == 'null':
            self.gp[type].covar.Cg.setCovariance(0.5 * self.covY)
            self.gp[type].covar.Cn.setCovariance(0.5 * self.covY)
        elif type == 'full':
            Cg0_K = self.gp['null'].covar.Cg.K()
            Cn0_K = self.gp['null'].covar.Cn.K()
            self.gp[type].covar.Cr.setCovariance((Cn0_K + Cg0_K) / 3.)
            self.gp[type].covar.Cg.setCovariance(2. * Cg0_K / 3.)
            self.gp[type].covar.Cn.setCovariance(2. * Cn0_K / 3.)
        elif type == 'block':
            Crf_K = self.gp['full'].covar.Cr.K()
            Cnf_K = self.gp['full'].covar.Cn.K()
            self.gp[type].covar.Cr.scale = sp.mean(Crf_K)
            self.gp[type].covar.Cn.setCovariance(Cnf_K)
        elif type == 'rank1':
            Crf_K = self.gp['full'].covar.Cr.K()
            Cnf_K = self.gp['full'].covar.Cn.K()
            self.gp[type].covar.Cr.setCovariance(Crf_K)
            self.gp[type].covar.Cn.setCovariance(Cnf_K)
        else:
            print('poppo')
        self.gp[type].optimize(factr=self.factr, verbose=False)
        RV = {'Cg': self.gp[type].covar.Cg.K(),
              'Cn': self.gp[type].covar.Cn.K(),
              'LML': sp.array([self.gp[type].LML()]),
              'LMLgrad': sp.array([sp.mean((self.gp[type].LML_grad()['covar'])**2)])}
        if type == 'null':
            RV['Cr'] = sp.zeros(RV['Cg'].shape)
        else:
            RV['Cr'] = self.gp[type].covar.Cr.K()
        if vc:
            # tr(P CoR) = tr(C)tr(R) - tr(Ones C) tr(Ones R) / float(NP)
            #           = tr(C)tr(R) - C.sum() * R.sum() / float(NP)
            trRr = (self.Xr**2).sum()
            var_r = sp.trace(RV['Cr']) * trRr / float(self.Y.size - 1)
            var_g = sp.trace(RV['Cg']) * self.trRg / float(self.Y.size - 1)
            var_n = sp.trace(RV['Cn']) * self.Y.shape[0]
            var_n -= RV['Cn'].sum() / float(RV['Cn'].shape[0])
            var_n /= float(self.Y.size - 1)
            RV['var'] = sp.array([var_r, var_g, var_n])
            if 0 and self.Y.size < 5000:
                pdb.set_trace()
                Kr = sp.kron(RV['Cr'], sp.dot(self.Xr, self.Xr.T))
                Kn = sp.kron(RV['Cn'], sp.eye(self.Y.shape[0]))
                _var_r = sp.trace(Kr - Kr.mean(0)) / float(self.Y.size - 1)
                _var_n = sp.trace(Kn - Kn.mean(0)) / float(self.Y.size - 1)
                _var = sp.array([_var_r, var_g, _var_n])
                print(((_var - RV['var'])**2).mean())
            if type == 'full':
                # calculate within region vcs
                Cr_block = sp.mean(RV['Cr']) * sp.ones(RV['Cr'].shape)
                Cr_rank1 = lowrank_approx(RV['Cr'], rank=1)
                var_block = sp.trace(Cr_block) * trRr / float(self.Y.size - 1)
                var_rank1 = sp.trace(Cr_rank1) * trRr / float(self.Y.size - 1)
                RV['var_r'] = sp.array(
                    [var_block, var_rank1 - var_block, var_r - var_rank1])
        return RV



r"""
****************************
Multi Trait Set Test (mtSet)
****************************

Set tests are a powerful approach for genome-wide association testing between
groups of genetic variants and quantitative traits.
Here we implement :class:`.MTSet`, a mixed-model approach that enables joint
analysis across multiple correlated traits while accounting for population
structure and relatedness.
:class:`.MTSet` effectively combines the benefits of set tests with multi-trait
modeling and is computationally efficient, enabling genetic analysis of large
cohorts (up to 500,000 individuals) and multiple traits.
A detailed description of the method can be found at [CRS15]_.

.. rubric:: References

.. [CRS15] Casale FP, Rakitsch B, Lippert C, Stegle O (2015) Efficient set tests for the genetic analysis of correlated traits. Nature Methods, Vol. 12, No. 8. (15 June 2015), pp. 755-758.

Public interface
^^^^^^^^^^^^^^^^
"""

from .mtset import MTSet

__all__ = ['MTSet']



from optparse import OptionParser
from limix.io import read_plink
import pandas as pd
import numpy as np


def sets_from_bim(
        bim,
        size=50000,
        step=None,
        chrom=None,
        minSnps=1,
        maxSnps=None):
    r"""
    Builds a dataframe of variant-sets from a bim considering
    a sliding window approach.

    Args:
        bim (:class:`pandas.DataFrame`):
            bim dataframe from ``:func:limix.io.read_plink``.
        size (int, optional):
            set size (in base pairs).
            The default value is 50000.
        step (int, optional):
            sliding-window step.
            The default value is ``size/2``.
        chrom (int, optional):
            can be set to extract set file for
            an only chromosome.
            By default a genome-wide set file
            is built.
        minSnps (int, optional):
            minimin number of variants.
            Sets with number of variants that is
            lower than ``minSnps`` are excluded.
            Default value is 1.
        maxSnps (int, optional):
            maximum number of variants.
            Default value is ``numpy.inf``.

    Returns:
        :class:`pandas.DataFrame`:
            dataframe of variant-sets. It has columns:

            - `"setid"`: window id
            - `"chrom"`: chromosome
            - `"start"`: start position
            - `"end"`: end position
            - `"nsnps"`: number of variants in the region
    """
    if step is None:
        step = int(0.5 * size)

    if maxSnps is None:
        maxSnps = np.inf

    if chrom is None:
        chroms = pd.unique(bim['chrom'])
    else:
        chroms = [chrom]

    out = []
    for _c in chroms:

        # calc start and end of regions
        pos = bim.query("chrom=='%s'" % _c)['pos'].values
        start = np.arange(pos.min(), pos.max(), step)
        end = start + size

        # filter based on minSnps
        nsnps = calc_nsnps(pos, start, end)
        Iok = (nsnps >= minSnps) & (nsnps <= maxSnps)
        start = start[Iok]
        end = end[Iok]
        nsnps = nsnps[Iok]

        # build array and append
        chrom = np.repeat(_c, start.shape[0])
        setid = np.array(['%s:%d-%d' % (_c, _s, _e)
                         for (_c, _s, _e) in zip(chrom, start, end)])
        _out = np.array([setid, chrom, start, end, nsnps], dtype=object).T
        out.append(_out)

    # concatenate and export
    out = np.concatenate(out)

    # convert to pandas dataframe and export
    out = pd.DataFrame(out, columns=['setid', 'chrom', 'start', 'end', 'nsnps'])
    return out


def annotate_sets(sets, bim, minSnps=1, maxSnps=None):
    r"""
    Helper function to annotate and filter variant-sets.

    Provided the variant-sets to consider and the bim of the bed
    file to analyse, it computes the number of variants within each set
    and filters them and returns a variant-set dataframe with all info.

    Args:
        sets (:class:`pandas.DataFrame`):
            dataframe defining the variant-sets.
            It should contain the columns:

            - `"setid"`: set id
            - `"chrom"`: chromosome
            - `"start"`: start position
            - `"end"`: end position

        bim (:class:`pandas.DataFrame`):
            bim dataframe from ``:func:limix.io.read_plink``.
        minSnps (int, optional):
            minimin number of variants.
            Sets with number of variants that is
            lower than ``minSnps`` are excluded.
            Default value is 1.
        maxSnps (int, optional):
            maximum number of variants.
            Default value is ``numpy.inf``.

    Returns:
        :class:`pandas.DataFrame`:
            dataframe of variant-sets. It has columns:

            - `"setid"`: set id
            - `"chrom"`: chromosome
            - `"start"`: start position
            - `"end"`: end position
            - `"nsnps"`: number of variants in the region
    """
    if maxSnps is None:
        maxSnps = np.inf

    out = []
    chrom = sets['chrom'].values
    uchroms = pd.unique(chrom)
    for _c in uchroms:

        # filter on chromosome
        pos = bim.query("chrom=='%s'" % _c)['pos'].values
        Ichr = chrom == _c
        setid = sets['setid'].values[Ichr]
        start = sets['start'].values[Ichr]
        end = sets['end'].values[Ichr]

        # filter based on minSnps
        nsnps = calc_nsnps(pos, start, end)
        Iok = (nsnps >= minSnps) & (nsnps <= maxSnps)
        setid = setid[Iok]
        start = start[Iok]
        end = end[Iok]
        nsnps = nsnps[Iok]

        # build array and append
        chrom = np.repeat(_c, start.shape[0])
        _out = np.array([setid, chrom, start, end, nsnps],
                        dtype=object).T
        out.append(_out)

    # concatenate and export
    out = np.concatenate(out)

    # convert to pandas dataframe and export
    out = pd.DataFrame(out, columns=['setid', 'chrom', 'start', 'end', 'nsnps'])
    return out


def calc_nsnps(pos, start, end):
    nsnps = np.zeros(start.shape[0], dtype=int)
    for i, r in enumerate(zip(start, end)):
        nsnps[i] = ((pos >= r[0]) & (pos < r[1])).sum()
    return nsnps


if __name__ == '__main__':

    import os

    bedpath = 'data/chrom22_subsample20_maf0.10'

    if not os.path.exists(bedpath + '.bim'):
        os.system('wget http://www.ebi.ac.uk/~casale/data.zip')
        os.system('unzip data.zip')

    (bim, fam, G) = read_plink(bedpath)

    # sets_from_bim
    sets = sets_from_bim(bim)

    # annotate_sets
    sets1 = annotate_sets(sets, bim, minSnps=100, maxSnps=200)



import sys
import scipy as SP


def vec(M):
    return M.reshape((M.size, 1), order='F')


def smartAppend(table, name, value):
    """
    helper function for apppending in a dictionary
    """
    if name not in list(table.keys()):
        table[name] = []
    table[name].append(value)


def smartSum(x, key, value):
    """ create a new page in x if key is not a page of x
        otherwise add value to x[key] """
    if key not in list(x.keys()):
        x[key] = value
    else:
        x[key] += value


def dumpDictHdf5(RV, o):
    """ Dump a dictionary where each page is a list or an array """
    for key in list(RV.keys()):
        o.create_dataset(
            name=key,
            data=SP.array(
                RV[key]),
            chunks=True,
            compression='gzip')


def smartDumpDictHdf5(RV, o):
    """ Dump a dictionary where each page is a list or an array or still a dictionary (in this case, it iterates)"""
    for key in list(RV.keys()):
        if isinstance(RV[key], dict):
            g = o.create_group(key)
            smartDumpDictHdf5(RV[key], g)
        else:
            o.create_dataset(
                name=key,
                data=SP.array(
                    RV[key]),
                chunks=True,
                compression='gzip')



r"""
*****************************
Variance Decomposition Module
*****************************

- :class:`.VarianceDecomposition`

Public interface
^^^^^^^^^^^^^^^^
"""

from .vardec import VarianceDecomposition

__all__ = ['VarianceDecomposition']



def read_gen(prefix):
    r"""
    Read GEN files into Pandas data frames.

    Parameters
    ----------
    prefix : str
        Path prefix to the set of GEN files.

    Returns
    -------
    sample : dask dataframe
    genotype : dask dataframe

    Examples
    --------
    .. doctest::

        >>> from limix.io import read_gen
        >>> from limix.io.examples import gen_file_example
        >>>
        >>> data = read_gen(gen_file_example())
        >>> print(data['sample']) #doctest: +NORMALIZE_WHITESPACE
          sample_id subject_id  missing  gender  age  age_of_onset  phenotype_1
        0       1A0       W001     0.00       2    4            -9            0
        1       1A1       W002     0.00       2    4            -9            0
        2       1A2       W003     0.00       2    4            -9            1
        3       1A3       W004     0.09       2    4            -9            1
        4       1A4       W005     0.00       2    4            -9            1
        >>> print(data['genotype'].head()) #doctest: +NORMALIZE_WHITESPACE
          snp_id  rs_id       pos alleleA alleleB 1A0       1A1       1A2       1A3  \
                                                   AA AB BB  AA AB BB  AA AB BB  AA
        0    SA1  rs001  10000000       A       G   0  0  1   0  0  1   0  0  1   0
        1    SA2  rs002  10010000       A       G   0  0  1   0  1  0   1  0  0   0
        2    SA3  rs003  10020000       C       T   1  0  0   0  1  0   0  0  1   0
        3    SA4  rs004  10030000       G       T   1  0  0   0  1  0   0  0  1   0
        4    SA5  rs005  10040000       C       G   0  0  1   0  1  0   1  0  0   0
        <BLANKLINE>
                          1A4
               AB      BB  AA      AB      BB
        0  0.4277  0.5721   0  0.0207  0.9792
        1  1.0000  0.0000   1  0.0000  0.0000
        2  0.9967  0.0000   0  0.0000  1.0000
        3  1.0000  0.0000   0  0.0000  1.0000
        4  1.0000  0.0000   1  0.0000  0.0000
    """

    from pandas import read_csv, MultiIndex

    df_sample = read_csv(prefix + '.sample', header=0, sep=' ', skiprows=[1])

    nsamples = df_sample.shape[0]

    col_level0_names = ['snp_id', 'rs_id', 'pos', 'alleleA', 'alleleB']
    col_level1_names = [''] * 5
    for s in df_sample['sample_id']:
        col_level0_names += [s] * 3
        col_level1_names += ['AA', 'AB', 'BB']

    tuples = list(zip(col_level0_names, col_level1_names))
    index = MultiIndex.from_tuples(tuples, names=['first', 'second'])

    df_gen = read_csv(prefix + '.gen', names=index, sep=' ')

    return dict(sample=df_sample, genotype=df_gen)



r"""
*************
limix package
*************

A flexible and fast mixed model toolbox.

"""

from __future__ import absolute_import as _absolute_import

import limix_core as core

from . import (heritability, io, iset, mtset, plot, qtl, scripts, stats, util,
               vardec)

__version__ = '1.0.16'


def test():
    import os
    p = __import__('limix').__path__[0]
    src_path = os.path.abspath(p)
    old_path = os.getcwd()
    os.chdir(src_path)

    try:
        return_code = __import__('pytest').main(['-q', '--doctest-modules'])
    finally:
        os.chdir(old_path)

    if return_code == 0:
        print("Congratulations. All tests have passed!")

    return return_code


__all__ = [
    'test', 'core', 'io', 'plot', 'qtl', 'stats', 'util', 'vardec', 'mtset',
    'iset', 'scripts', 'heritability', '__version__'
]



from optparse import OptionParser


def entry_point():
    from ..mtset.core.postprocessCore import postprocess

    parser = OptionParser()
    parser.add_option("--resdir", dest='resdir', type=str, default='./')
    parser.add_option("--outfile", dest='outfile', type=str, default=None)
    parser.add_option("--tol", dest='tol', type=float, default=4e-3)
    (options, args) = parser.parse_args()
    postprocess(options)



from optparse import OptionParser


def entry_point():
    from ..mtset.core.analyzeCore import analyze

    parser = OptionParser()
    parser.add_option("--bfile", dest='bfile', type=str, default=None)
    parser.add_option("--cfile", dest='cfile', type=str, default=None)
    parser.add_option("--pfile", dest='pfile', type=str, default=None)
    parser.add_option("--nfile", dest='nfile', type=str, default=None)
    parser.add_option("--wfile", dest='wfile', type=str, default=None)
    parser.add_option("--ffile", dest='ffile', type=str, default=None)
    parser.add_option("--resdir", dest='resdir', type=str, default='./')
    parser.add_option("--trait_idx", dest='trait_idx', type=str, default=None)

    parser.add_option("--rank_r", dest='rank_r', type=int, default=1)
    parser.add_option(
        "--colCovarType_r",
        dest='colCovarType_r',
        type=str,
        default='lowrank')

    # start window, end window and permutations
    parser.add_option("--start_wnd", dest='i0', type=int, default=None)
    parser.add_option("--end_wnd", dest='i1', type=int, default=None)
    parser.add_option("--perm", dest='perm_i', type=int, default=None)
    parser.add_option("--factr", dest='factr', type=float, default=1e7)

    parser.add_option(
        "--unique_variants",
        action="store_true",
        dest='unique_variants',
        default=False)

    parser.add_option(
        "--standardize",
        action="store_true",
        dest='standardize',
        default=False)

    (options, args) = parser.parse_args()

    analyze(options)



from os.path import dirname, realpath, join


def hdf5_file_example():
    dir_path = dirname(realpath(__file__))
    return join(dir_path, 'data', 'hdf5', 'data.h5')


def csv_file_example():
    dir_path = dirname(realpath(__file__))
    return join(dir_path, 'data', 'csv', 'data.csv')


def gen_file_example():
    dir_path = dirname(realpath(__file__))
    return join(dir_path, 'data', 'gen', 'example')


def numpy_kinship_file_example():
    dir_path = dirname(realpath(__file__))
    return join(dir_path, 'data', 'numpy', '1000G_kinship.npy')



import scipy as sp


def empirical_pvalues(xt, x0):
    r"""Function to compute empirical P values.

    Compute empirical P values from the test statistics
    observed on the data and the null test statistics
    (from permutations, parametric bootstraps, etc).

    Args:
        xt (array-like):
            test statistcs observed on data.
        x0 (array-like):
            null test statistcs.
            The minimum P value that can be
            estimated is ``1./float(len(x0))``.

    Returns:
        pv (ndarray):
            estimated empirical P values.

    Example
    -------

        .. doctest::

            >>> from numpy.random import RandomState
            >>> from limix.stats import empirical_pvalues
            >>> from numpy import set_printoptions
            >>> set_printoptions(4)
            >>> random = RandomState(1)
            >>>
            >>> xt = random.chisquare(1, 1000)
            >>> x0 = random.chisquare(1, 10000)
            >>>
            >>> pv = empirical_pvalues(xt, x0)
            >>>
            >>> print(pv.shape)
            (1000,)
            >>>
            >>> print(pv[:4]) # doctest: +SKIP
            [0.5599 1.     0.8389 0.7975]
    """
    idxt = sp.argsort(xt)[::-1]
    idx0 = sp.argsort(x0)[::-1]
    xts = xt[idxt]
    x0s = x0[idx0]
    it = 0
    i0 = 0
    _count = 0
    count = sp.zeros(xt.shape[0])
    while True:
        if x0s[i0] > xts[it]:
            _count += 1
            i0 += 1
            if i0 == x0.shape[0]:
                count[idxt[it:]] = _count
                break
        else:
            count[idxt[it]] = _count
            it += 1
            if it == xt.shape[0]:
                break
    pv = (count + 1) / float(x0.shape[0])
    pv[pv > 1.] = 1.
    return pv



r"""
**********
Statistics
**********

PCA
^^^

.. autofunction:: limix.stats.pca

Normalization
^^^^^^^^^^^^^

.. autofunction:: limix.stats.boxcox

P-values
^^^^^^^^

.. autofunction:: limix.stats.qvalues
.. autofunction:: limix.stats.empirical_pvalues

Kinship processing
^^^^^^^^^^^^^^^^^^

.. autofunction:: limix.stats.linear_kinship
.. autofunction:: limix.stats.gower_norm
.. autofunction:: limix.stats.indep_pairwise
.. autofunction:: limix.stats.maf

Chi2
^^^^

.. autofunction:: limix.stats.Chi2mixture

"""

from .chi2mixture import Chi2mixture
from .fdr import qvalues
from .kinship import gower_norm, linear_kinship
from .pca import pca
from .preprocess import indep_pairwise, maf
from .teststats import empirical_pvalues
from .trans import boxcox

__all__ = [
    'pca', 'boxcox', 'gower_norm', 'qvalues', 'empirical_pvalues',
    'Chi2mixture', 'indep_pairwise', 'maf', 'linear_kinship'
]



from __future__ import unicode_literals

import os
import sys

from setuptools import find_packages, setup

try:
    import pypandoc
    long_description = pypandoc.convert_file('README.md', 'rst')
except (OSError, IOError, ImportError):
    long_description = open('README.md').read()


def setup_package():
    src_path = os.path.dirname(os.path.abspath(sys.argv[0]))
    old_path = os.getcwd()
    os.chdir(src_path)
    sys.path.insert(0, src_path)

    needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
    pytest_runner = ['pytest-runner'] if needs_pytest else []

    setup_requires = ["cython", "numpy"] + pytest_runner
    install_requires = [
        'scikit-learn', 'limix-core', 'dask[complete]', 'h5py',
        'pandas-plink>=1.2.18', 'limix-legacy', 'glimix-core', 'joblib',
        'tqdm', 'scipy', 'distributed', 'numpy-sugar', 'ncephes', 'asciitree',
        'numpy', 'matplotlib'
    ]
    tests_require = ['pytest', 'pytest-console-scripts', 'pytest-pep8']

    console_scripts = [
        'limix_runner=limix.scripts.limix_runner:entry_point',
        'mtset_postprocess=limix.scripts.mtset_postprocess:entry_point',
        'mtset_preprocess=limix.scripts.mtset_preprocess:entry_point',
        'mtset_definesets=limix.scripts.mtset_definesets:entry_point',
        'mtset_simpheno=limix.scripts.mtset_simpheno:entry_point',
        'mtset_analyze=limix.scripts.mtset_analyze:entry_point',
        'limix_converter=limix.scripts.limix_converter:entry_point',
        'iset_analyze=limix.scripts.iset_analyze:entry_point',
        'iset_postprocess=limix.scripts.iset_postprocess:entry_point',
        'limix=limix.scripts.limix:entry_point',
        'ilimix=limix.scripts.ilimix:entry_point'
    ]

    metadata = dict(
        name='limix',
        version='1.0.16',
        maintainer="Limix Developers",
        maintainer_email="horta@ebi.ac.uk",
        author=("Christoph Lippert, Danilo Horta, " +
                "Francesco Paolo Casale, Oliver Stegle"),
        author_email="stegle@ebi.ac.uk",
        license="Apache License 2.0",
        description="A flexible and fast mixed model toolbox.",
        long_description=long_description,
        url='https://github.com/limix/limix',
        packages=find_packages(),
        zip_safe=False,
        install_requires=install_requires,
        setup_requires=setup_requires,
        tests_require=tests_require,
        include_package_data=True,
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        entry_points={
            'console_scripts': console_scripts
        })

    try:
        setup(**metadata)
    finally:
        del sys.path[0]
        os.chdir(old_path)


if __name__ == '__main__':
    setup_package()



from __future__ import division

from numpy import asarray, linspace


def plot_power_curve(df, color=None, ax=None):
    r"""Plot number of hits across significance levels.

    Parameters
    ----------

    df : :class:`pandas.DataFrame`
        Data frame with `pv` and `label` columns.
    color : dict
        Map colors to labels.
    ax : :class:`matplotlib.axes.AxesSubplot`
        The target handle for this figure. If None, the current axes is set.

    Returns
    -------
    :class:`matplotlib.axes.AxesSubplot`
        Axes.

    Examples
    --------

    .. plot::

        from limix.plot import plot_power_curve
        from pandas import DataFrame
        from numpy.random import RandomState
        from matplotlib import pyplot as plt
        random = RandomState(1)
        nsnps = 10000

        pv0 = list(random.rand(nsnps))
        pv1 = list(0.7 * random.rand(nsnps))

        fig = plt.figure(1, figsize=(5,5))
        plt.subplot(111)

        data = dict(pv=pv0 + pv1,
                    label=['label0'] * nsnps + ['label1'] * nsnps)
        df = DataFrame(data=data)
        plot_power_curve(df)
        plt.show()
    """

    import matplotlib.pyplot as plt

    ax = plt.gca() if ax is None else ax
    labels = list(df['label'].unique())

    if color is None:
        colors = _get_default_colors()
        color = {m: colors[i] for (i, m) in enumerate(labels)}

    alphas, nhits = _collect_nhits(df)

    for label in labels:
        ax.plot(
            alphas,
            asarray(nhits[label], int),
            color=color[label],
            label=label)

    _set_labels(ax)

    return ax


def _collect_nhits(df):
    labels = list(df['label'].unique())
    alphas = linspace(0.01, 0.5, 500)
    nhits = {l: [] for l in labels}

    for i in range(len(alphas)):
        alpha = alphas[i]
        for label in labels:
            ix = df['label'] == label
            df_ = df.loc[ix, :]
            ntests = df_.shape[0]
            n = (df_['pv'] < alpha).sum()
            nhits[label] += [n]

    for label in labels:
        nhits[label] = asarray(nhits[label], int)

    return (alphas, nhits)


def _set_labels(ax):
    ax.set_xlabel('significance level')
    ax.set_ylabel('number of hits')
    ax.legend()


def _get_default_colors():
    return ['red', 'green', 'blue']



import scipy as SP
import pdb
SP.random.seed(0)

if __name__ == '__main__':

    N = 500
    P = 4
    S = 1000
    R = 10

    hr = 0.05
    hg = 0.40

    Xr = 1. * (SP.rand(N, R) < 0.2)
    Xg = 1. * (SP.rand(N, S) < 0.2)
    Xg -= Xg.mean(0)
    Xg /= Xg.std(0)
    XX = SP.dot(Xg, Xg.T)
    XX /= XX.diagonal().mean()
    XX += 1e-4 * SP.eye(XX.shape[0])

    Yr = SP.dot(Xr, SP.randn(R, P))
    Yr *= SP.sqrt(hr / Yr.var(0))
    Yg = SP.dot(Xg, SP.randn(S, P))
    Yg *= SP.sqrt(hg / Yg.var(0))
    Yn = SP.randn(N, P)
    Yn *= SP.sqrt((1 - hr - hg) / Yn.var(0))

    Y = Yr + Yg + Yn
    Y -= Y.mean(0)
    Y /= Y.std(0)

    SP.savetxt('Xr.txt', Xr, fmt='%d')
    SP.savetxt('XX.txt', XX, fmt='%.6f')
    SP.savetxt('Y.txt', Y, fmt='%.6f')



import scipy as sp
import scipy.linalg as la
from .linalg_utils import lowrank_approx

ntype_dict = {'assoc': 'null', 'gxe': 'block', 'gxehet': 'rank1'}


def define_gp(Y, Xr, mean, Ie, type):
    from limix_core.covar import LowRankCov
    from limix_core.covar import FixedCov
    from limix_core.covar import FreeFormCov
    from limix_core.covar import CategoricalLR
    from limix_core.gp import GP
    P = 2
    if type == 'null':
        _Cr = FixedCov(sp.ones([2, 2]))
        _Cr.scale = 1e-9
        _Cr.act_scale = False
        covar = CategoricalLR(_Cr, sp.ones((Xr.shape[0], 1)), Ie)
    else:
        if type == 'block':
            _Cr = FixedCov(sp.ones((P, P)))
        elif type == 'rank1':
            _Cr = LowRankCov(P, 1)
        elif type == 'full':
            _Cr = FreeFormCov(P)
        else:
            print('poppo')
        covar = CategoricalLR(_Cr, Xr, Ie)
    _gp = GP(covar=covar, mean=mean)
    return _gp


class MvSetTestInc():

    def __init__(
            self,
            Y=None,
            Xr=None,
            F=None,
            factr=1e7,
            Ie=None,
            debug=False):
        """
        Args:
            Y:          [N, 1] phenotype matrix
            Xr:         [N, S] genotype data of the set component
            R:          [N, S] genotype data of the set component
            factr:      paramenter that determines the accuracy of the solution
                        (see scipy.optimize.fmin_l_bfgs_b for more details)
        """
        if F is None:
            F = sp.ones((y.shape[0], 1))
        # kroneckerize F
        W = sp.zeros((Y.shape[0], 2 * F.shape[1]))
        W[:, :F.shape[1]] = Ie[:, sp.newaxis] * F
        W[:, F.shape[1]:] = (~Ie[:, sp.newaxis]) * F
        from limix_core.mean import MeanBase
        self.mean = MeanBase(Y, W)
        # avoid SVD failus by adding some jitter
        Xr += 2e-6 * (sp.rand(*Xr.shape) - 0.5)
        # store stuff
        Xr -= Xr.mean(0)
        Xr /= Xr.std(0)
        Xr /= sp.sqrt(Xr.shape[1])
        self.Y = Y
        self.F = F
        self.Xr = Xr
        self.Ie = Ie
        self.covY = sp.cov(Y.T)
        self.factr = factr
        self.debug = debug
        self.gp = {}
        self.info = {}

    def assoc(self):
        # fit model
        for key in ['null', 'full']:
            if key not in list(self.gp.keys()):
                if self.debug:
                    print('.. dening %s' % key)
                self.gp[key] = define_gp(
                    self.Y, self.Xr, self.mean, self.Ie, key)
                if self.debug:
                    print('.. fitting %s' % key)
                self.info[key] = self._fit(key, vc=True)
        return self.info['null']['LML'] - self.info['full']['LML']

    def gxe(self):
        # fit model
        for key in ['null', 'full', 'block']:
            if key not in list(self.gp.keys()):
                if self.debug:
                    print('.. defining %s' % key)
                self.gp[key] = define_gp(
                    self.Y, self.Xr, self.mean, self.Ie, key)
                if self.debug:
                    print('.. fitting %s' % key)
                self.info[key] = self._fit(key, vc=True)
        return self.info['block']['LML'] - self.info['full']['LML']

    def gxehet(self):
        # fit model
        for key in ['null', 'full', 'rank1']:
            if key not in list(self.gp.keys()):
                if self.debug:
                    print('.. defining %s' % key)
                self.gp[key] = define_gp(
                    self.Y, self.Xr, self.mean, self.Ie, key)
                if self.debug:
                    print('.. fitting %s' % key)
                self.info[key] = self._fit(key, vc=True)
        return self.info['rank1']['LML'] - self.info['full']['LML']

    def assoc_null(self, n_nulls=30):
        LLR0 = sp.zeros(n_nulls)
        for ni in range(n_nulls):
            idx_perms = sp.random.permutation(self.Y.shape[0])
            _Xr = self.Xr[idx_perms]
            mvset0 = MvSetTestInc(Y=self.Y, F=self.F, Xr=_Xr, Ie=self.Ie)
            LLR0[ni] = mvset0.assoc()
        return LLR0

    def gxe_null(self, n_nulls=30):
        LLR0 = sp.zeros(n_nulls)
        for ni in range(n_nulls):
            Xb = sp.dot(self.mean.W, self.mean.b)
            Z = sp.randn(self.Y.shape[0], 1)
            _Y = Xb + self.gp['block'].covar.Kh_dot(Z)
            mvset0 = MvSetTestInc(Y=_Y, F=self.F, Xr=self.Xr, Ie=self.Ie)
            LLR0[ni] = mvset0.gxe()
        return LLR0

    def gxehet_null(self, n_nulls=30):
        LLR0 = sp.zeros(n_nulls)
        for ni in range(n_nulls):
            Xb = sp.dot(self.mean.W, self.mean.b)
            Z = sp.randn(self.Y.shape[0], 1)
            _Y = Xb + self.gp['rank1'].covar.Kh_dot(Z)
            mvset0 = MvSetTestInc(Y=_Y, F=self.F, Xr=self.Xr, Ie=self.Ie)
            LLR0[ni] = mvset0.gxehet()
        return LLR0

    def _fit(self, type, vc=False):
        # 2. init
        if type == 'null':
            self.gp[type].covar.Cn.setCovariance(sp.eye(2))
        elif type == 'full':
            Cr0_K = 1e-4 * sp.ones((2, 2)) + 1e-4 * sp.eye(2)
            Cn0_K = 0.99 * self.gp['null'].covar.Cn.K()
            self.gp[type].covar.Cr.setCovariance(Cr0_K)
            self.gp[type].covar.Cn.setCovariance(Cn0_K)
        elif type == 'block':
            Crf_K = self.gp['full'].covar.Cr.K()
            Cnf_K = self.gp['full'].covar.Cn.K()
            self.gp[type].covar.Cr.scale = sp.mean(Crf_K)
            self.gp[type].covar.Cn.setCovariance(Cnf_K)
        elif type == 'rank1':
            Crf_K = self.gp['full'].covar.Cr.K()
            Cnf_K = self.gp['full'].covar.Cn.K()
            self.gp[type].covar.Cr.setCovariance(Crf_K)
            self.gp[type].covar.Cn.setCovariance(Cnf_K)
        else:
            print('poppo')
        conv = self.gp[type].optimize(factr=self.factr, verbose=False)[0]
        B = self.gp[type].mean.b.reshape(
            (int(self.mean.W.shape[1] / 2), 2), order='F')
        RV = {'Cr': self.gp[type].covar.Cr.K(),
              'Cn': self.gp[type].covar.Cn.K(),
              'B': B,
              'conv': sp.array([conv]),
              'LML': sp.array([self.gp[type].LML()]),
              'LMLgrad': sp.array([sp.mean((self.gp[type].LML_grad()['covar'])**2)])}
        if vc:
            # tr(P WW) = tr(PWWP) = ((PW)**2).sum()
            # tr(P D) = (PD).sum() = D.sum() - 1/n * (Ones*D).sum()
            #                      = D.sum() - D.sum()
            def var_WW(W):
                PW = W - W.mean(0)
                rv = (PW**2).sum()
                rv /= float(W.shape[0] - 1)
                return rv
            var_r = var_WW(self.gp[type].covar.W())
            var_c = sp.var(sp.dot(self.mean.W, self.gp[type].mean.b))
            D = self.gp[type].covar.d_inv()**(-1)
            var_n = (1 - 1 / float(D.shape[0])) * \
                D.sum() / float(self.Y.size - 1)
            # var_n = sp.diagonal(sp.diag(D)-sp.diag(D).mean(0)).sum()/float(self.Y.size-1)
            RV['var'] = sp.array([var_r, var_c, var_n])
            if 0 and self.Y.size < 5000:
                pdb.set_trace()
                Kr = sp.kron(RV['Cr'], sp.dot(self.Xr, self.Xr.T))
                Kn = sp.kron(RV['Cn'], sp.eye(self.Y.shape[0]))
                _var_r = sp.trace(Kr - Kr.mean(0)) / float(self.Y.size - 1)
                _var_n = sp.trace(Kn - Kn.mean(0)) / float(self.Y.size - 1)
                _var = sp.array([_var_r, var_c, _var_n])
                print(((_var - RV['var'])**2).mean())
            if type == 'full':
                def build_W(C):
                    S, U = la.eigh(C)
                    I = S > 1e-9
                    Ch = U[:, I] * S[I]**0.5
                    RV = sp.zeros((self.Y.shape[0],
                                   Ch.shape[1] * self.Xr.shape[1]))
                    RV[self.Ie] = sp.kron(Ch[0], self.Xr[self.Ie])
                    RV[~self.Ie] = sp.kron(Ch[1], self.Xr[~self.Ie])
                    return RV
                # calculate within region vcs
                Cr_block = sp.mean(RV['Cr']) * sp.ones(RV['Cr'].shape)
                Cr_rank1 = lowrank_approx(RV['Cr'], rank=1)
                var_block = var_WW(build_W(Cr_block))
                var_rank1 = var_WW(build_W(Cr_rank1))
                # trRr = (self.Xr**2).sum()
                # var_block = sp.trace(Cr_block)*trRr / float(self.Y.size-1)
                # var_rank1 = sp.trace(Cr_rank1)*trRr / float(self.Y.size-1)
                RV['var_r'] = sp.array(
                    [var_block, var_rank1 - var_block, var_r - var_rank1])
        return RV



def entry_point():
    from optparse import OptionParser
    from ..mtset.core.iset_utils import calc_emp_pv_eff
    import pandas as pd
    import glob
    import os

    parser = OptionParser()
    parser.add_option("--resdir", dest='resdir', type=str, default='./')
    parser.add_option("--outfile", dest='outfile', type=str, default=None)
    # parser.add_option("--manhattan_plot", dest='manhattan',action="store_true",default=False)
    parser.add_option("--tol", dest='tol', type=float, default=4e-3)
    (options, args) = parser.parse_args()

    resdir = options.resdir
    out_file = options.outfile
    tol = options.tol

    print('.. load permutation results')
    file_name = os.path.join(resdir, '*.iSet.perm')
    files = glob.glob(file_name)
    df0 = pd.DataFrame()
    for _file in files:
        print(_file)
        df0 = df0.append(pd.read_csv(_file, index_col=0))

    print('.. load real results')
    file_name = os.path.join(resdir, '*.iSet.real')
    files = glob.glob(file_name)
    df = pd.DataFrame()
    for _file in files:
        print(_file)
        df = df.append(pd.read_csv(_file, index_col=0))

    # calculate P values for the three tests
    for test in ['mtSet', 'iSet', 'iSet-het']:
        df[test + ' pv'] = calc_emp_pv_eff(df[test + ' LLR'].values,
                                           df0[test + ' LLR0'].values)

    print(('.. saving %s' % out_file + '.res'))
    df.to_csv(out_file + '.res')



import scipy as SP
import pdb

if __name__ == '__main__':

    Cg_mtSet1VCNull = SP.loadtxt('./res_mtSet1VCnull_base_Cg.txt')
    Cn_mtSet1VCNull = SP.loadtxt('./res_mtSet1VCnull_base_Cn.txt')

    Cg_mtSetNull = SP.loadtxt('./res_mtSetNull_base_Cg.txt')
    Cn_mtSetNull = SP.loadtxt('./res_mtSetNull_base_Cn.txt')
    Cg_mtSetNullec = SP.loadtxt('./res_mtSetNull_eigenCache_Cg.txt')
    Cn_mtSetNullec = SP.loadtxt('./res_mtSetNull_eigenCache_Cn.txt')

    Cr_mtSet = SP.loadtxt('./res_mtSet_base_Cr.txt')
    Cg_mtSet = SP.loadtxt('./res_mtSet_base_Cg.txt')
    Cn_mtSet = SP.loadtxt('./res_mtSet_base_Cn.txt')
    Cr_mtSetec = SP.loadtxt('./res_mtSet_eigenCache_Cr.txt')
    Cg_mtSetec = SP.loadtxt('./res_mtSet_eigenCache_Cg.txt')
    Cn_mtSetec = SP.loadtxt('./res_mtSet_eigenCache_Cn.txt')

    # Cg_mtSet1VCNull = 0
    print(((Cg_mtSet1VCNull == 0).all()))

    # Eigenvalue chaching works
    print((((Cg_mtSetNull - Cg_mtSetNullec)**2 < 1e-4).all()))
    print((((Cn_mtSetNull - Cn_mtSetNullec)**2 < 1e-4).all()))
    print((((Cr_mtSet - Cr_mtSetec)**2 < 1e-4).all()))
    print((((Cg_mtSet - Cg_mtSetec)**2 < 1e-4).all()))
    print((((Cn_mtSet - Cn_mtSetec)**2 < 1e-4).all()))



import scipy as sp
import scipy.linalg as la


def msqrt(C):
    _U, _S = la.eigh(C)
    _S[_S < 0] = 0.
    return _U * sp.sqrt(_S)


def lowrank_approx(C, rank=1):
    _S, _U = la.eigh(C)
    _U = _U[:, -rank:]
    _S = _S[-rank:]
    R = sp.dot(_U * _S, _U.T)
    return R



def entry_point():
    import os
    import time
    from optparse import OptionParser

    import numpy as np
    import pandas as pd
    import scipy as sp

    from ..data import BedReader
    from ..iSet.iset import fit_iSet
    from ..util import unique_variants as f_uni_variants

    parser = OptionParser()
    parser.add_option("--bfile", dest='bfile', type=str, default=None)
    # parser.add_option("--cfile", dest='cfile', type=str, default=None)
    parser.add_option("--pfile", dest='pfile', type=str, default=None)
    parser.add_option("--wfile", dest='wfile', type=str, default=None)
    parser.add_option("--ffile", dest='ffile', type=str, default=None)
    parser.add_option("--ifile", dest='ifile', type=str, default=None)
    parser.add_option("--resdir", dest='resdir', type=str, default='./')

    # start window, end window and permutations
    parser.add_option("--n_perms", type=int, default=10)
    parser.add_option("--start_wnd", dest='i0', type=int, default=None)
    parser.add_option("--end_wnd", dest='i1', type=int, default=None)
    parser.add_option("--factr", dest='factr', type=float, default=1e7)

    parser.add_option(
        "--unique_variants",
        action="store_true",
        dest='unique_variants',
        default=False)

    parser.add_option(
        "--standardize",
        action="store_true",
        dest='standardize',
        default=False)

    (options, args) = parser.parse_args()

    print('importing data')
    F = sp.loadtxt(options.ffile + '.fe')
    Y = sp.loadtxt(options.pfile + '.phe')
    if len(Y.shape) == 1:
        Y = Y[:, sp.newaxis]

    sets = pd.DataFrame.from_csv(
        options.wfile + '.wnd', sep='\t', index_col=None)

    reader = BedReader(options.bfile)

    i0 = 1 if options.i0 is None else options.i0
    i1 = sets.shape[0] if options.i1 is None else options.i1

    df = pd.DataFrame()
    df0 = pd.DataFrame()

    if options.ifile is None:
        Ie = None
    else:
        Ie = sp.loadtxt(options.ifile + '.ind').flatten() == 1

    res_dir = options.resdir

    if not os.path.exists(res_dir):
        os.makedirs(res_dir)

    n_digits = len(str(sets.shape[0]))
    fname = str(i0).zfill(n_digits)
    fname += '_' + str(i1).zfill(n_digits)
    resfile = os.path.join(res_dir, fname)

    for wnd_i in range(i0, i1):
        t0 = time.time()

        _set = sets.ix[wnd_i]
        print('.. set %d: %s' % (wnd_i, _set['setid']))

        Xr = reader.getGenotypes(
            pos_start=_set['start'],
            pos_end=_set['end'],
            chrom=_set['chrom'],
            impute=True)

        if options.unique_variants:
            Xr = f_uni_variants(Xr)

        if options.standardize:
            Xr -= Xr.mean(0)
            Xr /= Xr.std(0)
        else:
            # encoding minor as 0
            p = 0.5 * Xr.mean(0)
            Xr[:, p > 0.5] = 2 - Xr[:, p > 0.5]

        Xr /= np.sqrt(Xr.shape[1])
        _df, _df0 = fit_iSet(Y, F=F, Xr=Xr, Ie=Ie, n_nulls=10)
        df = df.append(_df)
        df0 = df0.append(_df0)
        print('Elapsed:', time.time() - t0)

    df.to_csv(resfile + '.iSet.real')
    df0.to_csv(resfile + '.iSet.perm')



import os
import time
from optparse import OptionParser


class LIMIX_runner(object):
    '''
    A class for running a LIMIX pipeline.
    '''
    __slots__ = ["options", "args", "data", "result", "infostring"]

    def __init__(self, infostring=None):
        '''
        nothing to initialize
        '''
        self.options = None
        self.args = None
        self.data = None
        self.result = {}
        self.infostring = infostring
        if self.infostring is not None:
            self.result["infostring"] = self.infostring
        pass

    def parse_args(self):
        usage = "usage: %prog [options]"
        parser = OptionParser(usage=usage)
        parser.add_option(
            "-D",
            "--data_script",
            action="store",
            dest='data_script',
            type=str,
            help='A python script for loading the data',
            default=r"example_scripts/example_data.py")
        parser.add_option(
            "-E",
            "--experiment_script",
            action="store",
            dest='experiment_script',
            type=str,
            help='A python script for running the experiment',
            default=r"example_scripts/gwas_example.py")
        parser.add_option(
            "-O",
            "--outpath",
            action="store",
            dest='outpath',
            type=str,
            help='The output path (or file, if ending is .h5, the ooutput will be written to HDF5 file)',
            default="example_out")
        # parser.add_option('--seed', metavar='seed', type=int, help='The random seed', default=123123)
        parser.add_option(
            "-T",
            "--timestamp",
            action="store_true",
            dest='timestamp',
            help="Append unique timestamp to output value",
            default=False)
        parser.add_option(
            "-R",
            "--delimiter",
            action="store",
            dest='delimiter',
            type=str,
            help="The delimiter between output values",
            default="\t")
        parser.add_option(
            "-F",
            "--float_format",
            action="store",
            dest='float_format',
            type=str,
            help="Formating string for floating point output values",
            default="%.6e")
        (self.options, self.args) = parser.parse_args()
        self.result["options"] = str(self.options)
        return (self.options, self.args)

    def load_data(self):
        """
        Run the job specified in data_script
        """
        options = self.options
        command = open(self.options.data_script).read()
        self.result["data_script"] = command
        t0 = time.time()
        data = None  # fallback data
        exec(command)  # creates variable data
        t1 = time.time()
        print(("Elapsed time for data reading is %.2f seconds" % (t1 - t0)))
        self.data = data
        return self.data

    def run_experiment(self):
        """
        Run the job specified in experiment_script
        """
        data = self.data
        options = self.options
        result = self.result

        command = open(self.options.experiment_script).read()
        result["experiment_script"] = command
        t0 = time.time()
        exec(command)  # creates variable result
        t1 = time.time()
        print(("Elapsed time for running the experiment is %.2f seconds" %
               (t1 - t0)))
        self.result = result
        return self.result

    def write_resultfiles(self):
        """
        Write the output to disk
        """
        import limix_legacy.deprecated.io.output_writer as ow
        t0 = time.time()
        writer = ow.output_writer(output_dictionary=self.result)
        if len(self.options.outpath) >= 3 and self.options.outpath[
                -3:] == ".h5":
            writer.write_hdf5(
                filename=self.options.outpath,
                timestamp=self.options.timestamp)
        else:
            writer.write_txt(
                outdir=self.options.outpath,
                timestamp=self.options.timestamp,
                delimiter=self.options.delimiter,
                float_format=self.options.float_format)
        t1 = time.time()
        print(("Elapsed time for writing the output files is %.2f seconds" %
               (t1 - t0)))


def entry_point():
    infostring = "limix_runner.py, Copyright(c) 2014, The LIMIX developers\nlast modified: %s" % time.ctime(
        os.path.getmtime(__file__))
    print(infostring)

    runner = LIMIX_runner(infostring=infostring)
    (options, args) = runner.parse_args()
    data = runner.load_data()
    result = runner.run_experiment()
    runner.write_resultfiles()



# Copyright(c) 2014, The LIMIX developers (Christoph Lippert, Paolo Francesco Casale, Oliver Stegle)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# modified by Barbara Rakitsch
import pdb
import os
import numpy as SP
import csv


def standardize(snps):
    mdat = SP.ma.masked_array(snps, SP.isnan(snps))  # mask nan values
    # standardize column-wise to mean 0 and std 1
    N = (mdat - SP.mean(mdat, axis=0)) / SP.std(mdat, axis=0)
    return N.filled(0.0)  # assign 0.0 to masked missing values


def readBIM(basefilename, usecols=None):
    """
    helper method for speeding up read BED
    """
    bim = basefilename + '.bim'
    bim = SP.loadtxt(bim, dtype=bytes, usecols=usecols)
    return bim


def readFAM(basefilename, usecols=None):
    """
    helper method for speeding up read FAM
    """
    fam = basefilename + '.fam'
    fam = SP.loadtxt(fam, dtype=bytes, usecols=usecols)
    return fam


def readBED(
        basefilename,
        useMAFencoding=False,
        blocksize=1,
        start=0,
        nSNPs=SP.inf,
        startpos=None,
        endpos=None,
        order='F',
        standardizeSNPs=False,
        ipos=2,
        bim=None,
        fam=None):
    '''
    read [basefilename].bed,[basefilename].bim,[basefilename].fam
    --------------------------------------------------------------------------
    Input:
    basefilename    : string of the basename of [basename].bed, [basename].bim,
                      and [basename].fam
    blocksize       : load blocksize SNPs at a time (default 1)
    start           : index of the first SNP to be loaded from the .bed-file
                      (default 0)
    nSNPs           : load nSNPs from the .bed file (default SP.inf, meaning all)
    startpos        : starting position of the loaded genomic region[chr,bpdist]
    endpos          : end-position of the loaded genomic region     [chr,bpdist]
    order           : memory layout of the returned SNP array (default 'F')
                      'F'   : Fortran-style column-major array (SNP-major)
                      'C'   : C-style row-major array (individual-major)
    standardizeSNPs : bool indeicator if the resulting SNP array is supposed to
                      be zero-mean and unit-vatiance with mean imputed missing
                      values (default False)
    ipos            : the index of the position index to use (default 2)
                        1 : genomic distance
                        2 : base-pair distance
    useMAFencoding  : if set to one, the minor allele is encoded with 2, the major allele with 0.
                      otherwise, the plink coding is used (default False).
    --------------------------------------------------------------------------
    Output dictionary:
    'rs'     : [S] array rs-numbers
    'pos'    : [S*3] array of positions [chromosome, genetic dist, basepair dist]
    'snps'   : [N*S] array of snp-data
    'iid'    : [N*2] array of family IDs and individual IDs
    --------------------------------------------------------------------------
    '''

    if bim is None:
        bim = readBIM(basefilename, usecols=(0, 1, 2, 3))
    if fam is None:
        fam = readFAM(basefilename, usecols=(0, 1))

    rs = bim[:, 1]
    pos = SP.array(bim[:, (0, 2, 3)], dtype='float')

    if startpos is not None:
        # pdb.set_trace()
        i_c = pos[:, 0] == startpos[0]
        i_largerbp = pos[:, ipos] >= startpos[ipos]
        start = which(i_c * i_largerbp)
        while (start - 1 >= 0 and pos[start - 1, ipos] == startpos[ipos]):
            start = start - 1
        i_c = pos[:, 0] == endpos[0]
        i_smallerbp = pos[:, ipos] >= endpos[ipos]
        end = which(i_c * i_smallerbp)
        while (end + 1 < pos.shape[0] and pos[end + 1, ipos] == endpos[ipos]):
            end = end + 1
        nSNPs = end - start
        if (nSNPs <= 0) or (end == 0) or (start <= 0):
            ret = {
                'pos': SP.zeros((0, 3)),
                'rs': SP.zeros((0)),
                'iid': fam,
                'snps': SP.zeros((fam.shape[0], 0))
            }
            return ret
        pass
    N = fam.shape[0]
    S = bim.shape[0]
    S_res = min(S, start + nSNPs)
    nSNPs = min(S - start, nSNPs)
    # if startpos is not None:
    # print("start: " + str(start))
    # print("end: " + str(end))
    # print("S_res: " + str(S_res))
    # print("nSNPs: " + str(nSNPs))
    if nSNPs <= 0:
        ret = {
            'rs': rs[start:start],
            'pos': pos[start:start, :],
            # 'snps'   :SNPs[0:N,start:start],
            'snps': SP.zeros((N, 0)),
            'iid': fam
        }
        return ret
    SNPs = SP.zeros((int(SP.ceil(0.25 * N) * 4), nSNPs), order=order)
    bed = basefilename + '.bed'
    with open(bed, "rb") as f:
        mode = f.read(2)
        if mode != b'l\x1b':
            raise Exception('No valid binary PED file')
        mode = f.read(1)  # \x01 = SNP major \x00 = individual major
        if mode != b'\x01':
            raise Exception('only SNP-major is implemented')
        startbit = int(SP.ceil(0.25 * N)) * start + 3
        f.seek(int(startbit))
        for blockStart in SP.arange(0, nSNPs, blocksize, dtype=int):
            blockEnd = int(min(S, blockStart + blocksize))
            Sblock = min(nSNPs - blockStart, blocksize)
            nbyte = int(SP.ceil(0.25 * N) * Sblock)
            bytes = SP.array(bytearray(f.read(nbyte))).reshape(
                (int(SP.ceil(0.25 * N)), Sblock), order='F')

            SNPs[3::4, blockStart:blockEnd][bytes >= 64] = SP.nan
            SNPs[3::4, blockStart:blockEnd][bytes >= 128] = 1
            SNPs[3::4, blockStart:blockEnd][bytes >= 192] = 2
            bytes = SP.mod(bytes, 64)
            SNPs[2::4, blockStart:blockEnd][bytes >= 16] = SP.nan
            SNPs[2::4, blockStart:blockEnd][bytes >= 32] = 1
            SNPs[2::4, blockStart:blockEnd][bytes >= 48] = 2
            bytes = SP.mod(bytes, 16)
            SNPs[1::4, blockStart:blockEnd][bytes >= 4] = SP.nan
            SNPs[1::4, blockStart:blockEnd][bytes >= 8] = 1
            SNPs[1::4, blockStart:blockEnd][bytes >= 12] = 2
            bytes = SP.mod(bytes, 4)
            SNPs[0::4, blockStart:blockEnd][bytes >= 1] = SP.nan
            SNPs[0::4, blockStart:blockEnd][bytes >= 2] = 1
            SNPs[0::4, blockStart:blockEnd][bytes >= 3] = 2

    if 0:  # the binary format as described in the documentation (seems wrong)
        SNPs[3::4][bytes >= 128] = SP.nan
        SNPs[3::4][bytes >= 192] = 1
        bytes = SP.mod(bytes, 128)
        SNPs[3::4][bytes >= 64] += 1
        bytes = SP.mod(bytes, 64)
        SNPs[2::4][bytes >= 32] = SP.nan
        SNPs[2::4][bytes >= 48] = 1
        bytes = SP.mod(bytes, 32)
        SNPs[2::4][bytes >= 16] += 1
        bytes = SP.mod(bytes, 16)
        SNPs[1::4][bytes >= 8] = SP.nan
        SNPs[1::4][bytes >= 12] = 1
        bytes = SP.mod(bytes, 8)
        SNPs[1::4][bytes >= 4] += 1
        bytes = SP.mod(bytes, 4)
        SNPs[0::4][bytes >= 2] = SP.nan
        SNPs[0::4][bytes >= 3] = 1
        bytes = SP.mod(bytes, 2)
        SNPs[0::4][bytes >= 1] += 1
    snps = SNPs[0:N, :]

    if useMAFencoding:
        imaf = SP.sum(snps == 2, axis=0) > SP.sum(snps == 0, axis=0)
        snps[:, imaf] = 2 - snps[:, imaf]

    if standardizeSNPs:
        snps = standardize(snps)
    ret = {
        'rs': rs[start:S_res],
        'pos': pos[start:S_res, :],
        'snps': snps,
        'iid': fam
    }
    return ret



from os.path import exists


def file_type(filepath):
    imexts = ['.png', '.bmp', '.jpg', 'jpeg']
    if filepath.endswith('.hdf5') or filepath.endswith('.h5'):
        return 'hdf5'
    if filepath.endswith('.csv'):
        return 'csv'
    if filepath.endswith('.grm.raw'):
        return 'grm.raw'
    if filepath.endswith('.npy'):
        return 'npy'
    if _is_bed(filepath):
        return 'bed'
    if any([filepath.endswith(ext) for ext in imexts]):
        return 'image'
    return 'unknown'


def _is_bed(filepath):
    return all([exists(filepath + ext) for ext in ['.bed', '.bim', '.fam']])



import scipy as SP
import os


def readBimFile(basefilename):
    """
    Helper fuinction that reads bim files
    """
    # read bim file
    bim_fn = basefilename + '.bim'
    rv = SP.loadtxt(bim_fn, delimiter='\t', usecols=(0, 3), dtype=int)
    return rv


def readCovarianceMatrixFile(cfile, readCov=True, readEig=True):
    """"
    reading in similarity matrix

    cfile   File containing the covariance matrix. The corresponding ID file must be specified in cfile.id)
    """
    covFile = cfile + '.cov'
    evalFile = cfile + '.cov.eval'
    evecFile = cfile + '.cov.evec'

    RV = {}
    if readCov:
        assert os.path.exists(covFile), '%s is missing.' % covFile
        RV['K'] = SP.loadtxt(covFile)
    if readEig:
        assert os.path.exists(evalFile), '%s is missing.' % evalFile
        assert os.path.exists(evecFile), '%s is missing.' % evecFile
        RV['eval'] = SP.loadtxt(evalFile)
        RV['evec'] = SP.loadtxt(evecFile)

    return RV


def readCovariatesFile(fFile):
    """"
    reading in covariate file

    cfile   file containing the fixed effects as NxP matrix
            (N=number of samples, P=number of covariates)
    """
    assert os.path.exists(fFile), '%s is missing.' % fFile
    F = SP.loadtxt(fFile)
    if F.ndim == 1:
        F = F[:, SP.newaxis]
    return F


def readPhenoFile(pfile, idx=None):
    """"
    reading in phenotype file

    pfile   root of the file containing the phenotypes as NxP matrix
            (N=number of samples, P=number of traits)
    """

    usecols = None
    if idx is not None:
        """ different traits are comma-seperated """
        usecols = [int(x) for x in idx.split(',')]

    phenoFile = pfile + '.phe'
    assert os.path.exists(phenoFile), '%s is missing.' % phenoFile
    Y = SP.loadtxt(phenoFile, usecols=usecols)
    if (usecols is not None) and (len(usecols) == 1):
        Y = Y[:, SP.newaxis]

    Y -= Y.mean(0)
    Y /= Y.std(0)
    return Y


def readNullModelFile(nfile):
    """"
    reading file with null model info

    nfile   File containing null model info
    """

    params0_file = nfile + '.p0'
    nll0_file = nfile + '.nll0'
    assert os.path.exists(params0_file), '%s is missing.' % params0_file
    assert os.path.exists(nll0_file), '%s is missing.' % nll0_file
    params = SP.loadtxt(params0_file)
    NLL0 = SP.array([float(SP.loadtxt(nll0_file))])

    if params.ndim == 1:
        rv = {'params0_g': SP.array([params[0]]),
              'params0_n': SP.array([params[1]]),
              'NLL0': NLL0}
    else:
        rv = {'params0_g': params[0], 'params0_n': params[1], 'NLL0': NLL0}

    return rv


def readWindowsFile(wfile):
    """"
    reading file with windows

    wfile   File containing window info
    """
    window_file = wfile + '.wnd'
    assert os.path.exists(window_file), '%s is missing.' % window_file
    rv = SP.loadtxt(window_file)
    return rv



from __future__ import division

import scipy.spatial
from joblib import Parallel, cpu_count, delayed
from numpy import (asarray, ascontiguousarray, double, einsum, isfinite,
                   logical_not, minimum, newaxis, sqrt, unique, zeros)
from scipy.spatial import _distance_wrap
from scipy.spatial.distance import pdist
from tqdm import tqdm


def _row_norms(X):
    norms = einsum('ij,ij->i', X, X, dtype=double)
    return sqrt(norms, out=norms)


def _sq_pearson(X):
    m = X.shape[0]
    dm = zeros((m * (m - 1)) // 2, dtype=double)

    X2 = X - X.mean(1)[:, newaxis]
    X2 = ascontiguousarray(X2)
    dm = pdist(X2, 'cosine')
    # norms = _row_norms(X2)
    # _distance_wrap.pdist_cosine_wrap(X2, dm, norms)
    return (-dm + 1)**2


def _pdist_threshold(mark, dist, thr):
    mark[:] = False
    size = len(mark)

    l = 0
    for i in range(0, size - 1):
        if mark[i]:
            l += size - (i + 1)
            continue

        for j in range(i + 1, size):
            if dist[l] > thr:
                mark[j] = True
            l += 1


def func(x, excls, threshold):
    dist = _sq_pearson(x)
    e = zeros(x.shape[0], dtype=bool)
    _pdist_threshold(e, dist, threshold)
    excls |= e


def indep_pairwise(X, window_size, step_size, threshold, verbose=True):
    r"""
    Determine pair-wise independent variants.

    Independent variants are defined via squared Pearson correlations between
    pairs of variants inside a sliding window.

    Parameters
    ----------
    X : array_like
        Sample by variants matrix.
    window_size : int
        Number of variants inside each window.
    step_size : int
        Number of variants the sliding window skips.
    threshold : float
        Squared Pearson correlation threshold for independence.
    verbose : bool
        `True` for progress information; `False` otherwise.

    Returns
    -------
    ok : boolean array defining independent variants

    Examples
    --------
    .. doctest::

        >>> from numpy.random import RandomState
        >>> from limix.stats import indep_pairwise
        >>>
        >>> random = RandomState(0)
        >>> X = random.randn(10, 20)
        >>>
        >>> indep_pairwise(X, 4, 2, 0.5, verbose=False) # doctest: +SKIP
        array([ True,  True, False,  True,  True,  True,  True,  True,  True,
                True,  True,  True,  True,  True,  True,  True,  True,  True,
                True,  True])
    """
    left = 0
    excls = zeros(X.shape[1], dtype=bool)
    excl = zeros(window_size, dtype=bool)

    assert step_size <= window_size

    n = (X.shape[1] + step_size) // step_size

    steps = list(range(n))
    cc = max(1, cpu_count())

    with tqdm(total=n, desc='Indep. pairwise', disable=not verbose) as pbar:

        while len(steps) > 0:
            i = 0
            right = 0
            delayeds = []
            while i < len(steps):

                step = steps[i]
                left = step * step_size
                if left < right:
                    i += 1
                    continue

                del steps[i]
                right = min(left + window_size, X.shape[1])
                x = ascontiguousarray(X[:, left:right].T)

                delayeds.append(delayed(func)(x, excls[left:right], threshold))
                if len(delayeds) == cc:
                    Parallel(
                        n_jobs=min(len(delayeds), cc),
                        backend='threading')(delayeds)
                    pbar.update(len(delayeds))
                    delayeds = []

            if len(delayeds) == 0:
                continue

            Parallel(
                n_jobs=min(len(delayeds), cc), backend='threading')(delayeds)
            pbar.update(len(delayeds))

    return logical_not(excls)


def _check_encoding(X):
    u = unique(X)
    u = u[isfinite(u)]
    if len(u) > 3:
        return False
    return all([i in set([0, 1, 2]) for i in u])


def maf(X):
    r"""Compute minor allele frequencies.

    It assumes that `X` encodes 0, 1, and 2 representing the number
    of alleles.

    Args:
        X (array_like): Genotype matrix.

    Returns:
        array_like: minor allele frequencies.

    Examples
    --------

        .. doctest::

            >>> from numpy.random import RandomState
            >>> from limix.stats import maf
            >>>
            >>> random = RandomState(0)
            >>> X = random.randint(0, 3, size=(100, 10))
            >>>
            >>> print(maf(X)) # doctest: +SKIP
            [0.49  0.49  0.445 0.495 0.5   0.45  0.48  0.48  0.47  0.435]
    """
    ok = _check_encoding(X)
    if not ok:
        raise ValueError("It assumes that X encodes 0, 1, and 2 only.")
    s0 = X.sum(0)
    s0 = s0 / (2 * X.shape[0])
    s1 = 1 - s0
    return minimum(s0, s1)



from numpy import loadtxt


def read_plink(prefix, verbose=True):
    r"""
    Read PLINK files into Pandas data frames.

    Parameters
    ----------
    prefix : str
        Path prefix to the set of PLINK files.
    verbose : bool
        `True` for progress information; `False` otherwise.

    Returns
    -------
    alleles : pandas dataframe
    samples : pandas dataframe
    genotype : ndarray

    Examples
    --------
    .. doctest::

        >>> from limix.io import read_plink
        >>> from pandas_plink import example_file_prefix
        >>>
        >>> (bim, fam, bed) = read_plink(example_file_prefix(), verbose=False)
        >>> print(bim.head()) #doctest: +NORMALIZE_WHITESPACE
          chrom         snp   cm    pos a0 a1  i
        0     1  rs10399749  0.0  45162  G  C  0
        1     1   rs2949420  0.0  45257  C  T  1
        2     1   rs2949421  0.0  45413  0  0  2
        3     1   rs2691310  0.0  46844  A  T  3
        4     1   rs4030303  0.0  72434  0  G  4
        >>> print(fam.head()) #doctest: +NORMALIZE_WHITESPACE
                fid       iid    father    mother gender trait  i
        0  Sample_1  Sample_1         0         0      1    -9  0
        1  Sample_2  Sample_2         0         0      2    -9  1
        2  Sample_3  Sample_3  Sample_1  Sample_2      2    -9  2
        >>> print(bed.compute()) # doctest: +SKIP
        [[ 2.  2.  1.]
         [ 2.  1.  2.]
         [nan nan nan]
         [nan nan  1.]
         [ 2.  2.  2.]
         [ 2.  2.  2.]
         [ 2.  1.  0.]
         [ 2.  2.  2.]
         [ 1.  2.  2.]
         [ 2.  1.  2.]]

    Notice the ``i`` column in bim and fam data frames. It maps to the
    corresponding position of the bed matrix:

    .. doctest::

        >>> from limix.io import read_plink
        >>> from pandas_plink import example_file_prefix
        >>>
        >>> (bim, fam, bed) = read_plink(example_file_prefix(), verbose=False)
        >>> chrom1 = bim.query("chrom=='1'")
        >>> X = bed[chrom1.i.values, :].compute()
        >>> print(X) #doctest: +NORMALIZE_WHITESPACE +SKIP
        [[ 2.  2.  1.]
         [ 2.  1.  2.]
         [nan nan nan]
         [nan nan  1.]
         [ 2.  2.  2.]
         [ 2.  2.  2.]
         [ 2.  1.  0.]
         [ 2.  2.  2.]
         [ 1.  2.  2.]
         [ 2.  1.  2.]]
    """
    from pandas_plink import read_plink
    return read_plink(prefix, verbose=verbose)


def read_grm_raw(filepath):
    return loadtxt(filepath)


def see_kinship(filepath):
    import limix

    if filepath.endswith('.grm.raw'):
        K = read_grm_raw(filepath)

    limix.plot.plot_kinship(K)


def _print_title(title, msg):
    k = msg.find("\n") - len(title) - 2
    left = ("-" * (k // 2)) + " "
    right = " " + ("-" * (k // 2 + k % 2))
    print(left + title + right)
    print(msg)


def see_bed(filepath):
    (bim, fam, bed) = read_plink(filepath)

    _print_title("Samples", repr(bim))
    _print_title("Genotype", repr(fam))
